# -*- coding: utf-8 -*-

"""package ad9xdds
author    Benoit Dubois
copyright FEMTO ENGINEERING, 2021-2025
license   GPL v3.0+
brief     Emulation of basic DDS board AD9912Dev
"""

import logging
import inspect
import signalslot as ss
from ad9xdds.dds_core import AbstractDds


AD_VENDOR_ID = 0x0456
AD9912DEV_PRODUCT_ID = 0xee09
AD9549DEV_PRODUCT_ID = 0xee08
DDS_DEVICE_LIST = {'AD9912': 'Ad9912Dev', 'AD9915': 'Ad9915DevUmr232Hm'}


# =============================================================================
class Ad9912Emul(AbstractDds):
    """Class 'emulating' the AD9912 DDS development board.
    Used for test when no real device is available.
    """

    FTW_SIZE = 48               # Frequency Tuning Word register size (bit)
    PHASE_SIZE = 14             # Phase register size (bit)
    DAC_OUT_SIZE = 10           # Output DAC resolution (bit)
    IFMAX = 1000000000.0        # Input maximum frequency (Hz)
    OFMAX = 400000000.0         # Output maximum frequency (Hz)
    AMAX = (1 << DAC_OUT_SIZE) - 1

    def __init__(self, ifreq=IFMAX):
        super().__init__()
        self._ifreq = None
        self._ofreq = None
        self._phy = None
        self._amp = None
        self._is_connected = False
        logging.info("Init DDS test device: %r", self)

    def connect(self, vendor_id=None, product_id=None,
                bus=None, address=None):
        logging.info("Connect to DDS test device")
        self._is_connected = True
        return True

    def disconnect(self):
        logging.debug("Disconnection")

    def is_connected(self):
        return self._is_connected

    def set_ifreq(self, ifreq):
        logging.debug("DdsTestDev.set_ifreq(" + str(ifreq) + ")")
        self._ifreq = float(ifreq)
        return ifreq

    def get_ifreq(self):
        logging.debug("DdsTestDev.get_ifreq() = " + str(self._ifreq))
        return self._ifreq

    def set_ofreq(self, ofreq):
        logging.info("DdsTestDev.set_ofreq(" + str(ofreq) + ")")
        self._ofreq = float(ofreq)
        return ofreq

    def get_ofreq(self):
        logging.debug("DdsTestDev.get_ofreq() = %r", self._ofreq)
        return self._ofreq

    def set_phy(self, phy):
        logging.debug("DdsTestDev.set_phy(" + str(phy) + ")")
        self._phy = phy
        return phy

    def get_phy(self):
        logging.debug("DdsTestDev.get_phy() = " + str(self._phy))
        return self._phy

    def set_amp(self, fsc):
        logging.debug("DdsTestDev.set_amp(" + str(fsc) + ")")
        self._amp = fsc
        return fsc

    def get_amp(self):
        logging.debug("DdsTestDev.get_amp() = " + str(self._amp))
        return self._amp

    def set_hstl_output_state(self, state=False):
        logging.debug("Set HSTL output state to: " + str(state))

    def get_hstl_output_state(self):
        pass

    def set_cmos_output_state(self, state=False):
        logging.debug("Set CMOS output state to: " + str(state))

    def get_cmos_output_state(self):
        pass

    def set_pll_state(self, state=False):
        logging.debug("Set PLL state to: %s", str(state))

    def get_pll_state(self):
        pass

    def set_cp_current(self, value=0):
        logging.debug("Set charge pump current to: " + str(value))

    def get_cp_current(self):
        pass

    def set_vco_range(self, value=None):
        logging.debug("Set VCO range to: " + str(value))

    def get_vco_range(self):
        pass

    def set_hstl_doubler_state(self, flag=False):
        logging.debug("Set HSTL doubler state to: " + str(flag))

    def get_hstl_doubler_state(self):
        pass

    def set_pll_doubler_state(self, flag=False):
        logging.debug("Set PLL doubler state to: " + str(flag))

    def get_pll_doubler_state(self):
        pass

    def set_pll_multiplier_factor(self, factor):
        logging.debug("Set PLL multiplier factor to: " + str(factor))

    def get_pll_multiplier_factor(self):
        pass

    def set_led(self, flag=False):
        logging.debug("Set LED blink: " + str(flag))

    def set_reg(self, address, value):
        logging.debug("Set " + str(value) + " @adress " + str(address))

    def get_reg(self, address):
        logging.debug("Request value @adress " + str(address))


# =============================================================================
class SAd9912Emul(Ad9912Emul):
    """Class derived from Ad9912Emul class to add signal/slot facilities.
    """

    def __init__(self, **kwargs):
        self.ifreq_updated = ss.Signal(['value'])
        self.ofreq_updated = ss.Signal(['value'])
        self.phase_updated = ss.Signal(['value'])
        self.amp_updated = ss.Signal(['value'])
        self.pll_state_updated = ss.Signal(['flag'])
        self.pll_doubler_updated = ss.Signal(['flag'])
        self.pll_factor_updated = ss.Signal(['value'])
        super().__init__(**kwargs)

    def connect(self, **kwargs):
        return super().connect(**kwargs)

    def set_ifreq(self, value, **kwargs):
        super().set_ifreq(value)
        self.ifreq_updated.emit(value=value)

    def set_ofreq(self, value, **kwargs):
        aofreq = super().set_ofreq(value)
        self.ofreq_updated.emit(value=aofreq)
        return aofreq

    def set_phy(self, value, **kwargs):
        aphy = super().set_phy(value)
        self.phase_updated.emit(value=aphy)
        return aphy

    def set_amp(self, value, **kwargs):
        aamp = super().set_amp(value)
        self.amp_updated.emit(value=aamp)
        return aamp

    def set_pll_state(self, flag, **kwargs):
        super().set_pll_state(flag)
        self.pll_state_updated.emit(flag=flag)

    def set_pll_doubler_state(self, flag, **kwargs):
        super().set_pll_doubler_state(flag)
        self.pll_doubler_updated.emit(flag=flag)

    def set_pll_multiplier_factor(self, value, **kwargs):
        super().set_pll_multiplier_factor(value)
        self.pll_factor_updated.emit(value=value)


# =============================================================================
class Ad9915DevUmr232HmEmul():

    def __init__(self, ifreq=2500000000):
        """The constructor.
        :param ifreq: Current input frequency in Hz (float)
        :returns: None
        """
        self._ifreq = ifreq
        self._ofreq = 8e6
        self._phy = 0
        self._amp = 511
        self._profile = 0
        self._is_connected = False
        logging.debug("DdsTestDev.%r(%r) done", inspect.stack()[0][3], ifreq)

    def connect(self, url):
        """Connection process.
        :param url: FTDI url like 'ftdi://ftdi:232h:FT0GPCDF/0' (str)
        :returns: None
        """
        self._is_connected = True
        logging.debug("DdsTestDev.%r(%r)", inspect.stack()[0][3], url)

    def disconnect(self):
        self._is_connected = False
        logging.debug("DdsTestDev.%r()", inspect.stack()[0][3])

    def is_connected(self):
        """Return True if interface to DDS board is ready else return False.
        """
        logging.debug("DdsTestDev.%r()", inspect.stack()[0][3])
        if self.is_connected:
            return True
        return False

    def dac_calibration(self):
        """DAC calibration, needed after each power-up and every time REF CLK or
        the internal system clock is changed.
        :returns: None
        """
        logging.debug("DdsTestDev.%r()", inspect.stack()[0][3])

    def set_profile(self, profile):
        """Set profile currently in use. Curently not implemented.
        :param profile: Select profile in use between 0 to 7 (int)
        :returns: None
        """
        if profile not in range(0, 8):
            raise ValueError("Profile must be in range 0 to 7, here: %r",
                             profile)
        self._profile = profile

    def get_profile(self):
        """Get profile currently in use. Curently not implemented.
        :returns: profile currently in use (int)
        """
        return self._profile

    def set_operation_mode(self, mode):
        pass

    def get_sysfreq(self):
        """Get system frequency.
        Currently, does not support PLL.
        :returns: Current system frequency value (float)
        """
        '''if self._pll_state is True:
            if self._pll_doubler_state is True:
                doubler = 2.0
            else:
                doubler = 1.0
            factor = self._pll_factor
            sfreq = self._ifreq * doubler * factor
        else:
            sfreq = self._ifreq
        return sfreq
        '''
        return self._ifreq

    def set_ifreq(self, value):
        """Set input frequency.
        :param value: Input frequency value (float)
        :returns: None
        """
        self._ifreq = value
        self.set_ofreq(self._ofreq)

    def get_ifreq(self):
        """Get input frequency.
        :returns: Current input frequency value (float)
        """
        return self._ifreq

    def set_ofreq_fine(self, ofreq):
        if self.is_connected() is False:
            logging.error("Device is not connected.")
            return None
        self._ofreq = value
        logging.debug("DdsTestDev.%r(%r, %r)", inspect.stack()[0][3], value, profile)
        return self._ofreq

    def set_ofreq(self, value, profile=None):
        """Set output frequency to current DDS profile if profile parameter is
        None or set output frequency of requested DDS profile.
        Return the actual output frequency (see _actual_ofreq() method).
        :param value: Output frequency value (float).
        :param profile: Profile to update between 0 to 7 (int)
        :returns: Actual output frequency (float)
        """
        if self.is_connected() is False:
            logging.error("Device is not connected.")
            return None
        self._ofreq = value
        logging.debug("DdsTestDev.%r(%r, %r)", inspect.stack()[0][3], value, profile)
        return self._ofreq

    def get_ofreq(self, profile=None):
        """Get output frequency of current DDS profile if profile parameter is
        None or return output frequency of requested DDS profile.
        :param profile: Profile output frequency requested (int)
        :returns: Output frequency of DDS profile (float).
        """
        if self.is_connected() is False:
            logging.error("Device is not connected.")
            return None
        logging.debug("DdsTestDev.%r(%r)", inspect.stack()[0][3], profile)
        return self._ofreq

    def set_phy(self, value, profile=None):
        """Set phase of output signal on DDS.
        Take the queried output phase (in degree) as argument and set
        the adequat register in the DDS.
        :param value: Output phase value (float).
        :param profile: Profile to update between 0 to 7 (int)
        :returns: Actual output phase (float)
        """
        if self.is_connected() is False:
            logging.error("Device is not connected.")
            return None
        self._phy = value
        logging.debug("DdsTestDev.%r(%r, %r)", inspect.stack()[0][3], value, profile)
        return 0

    def get_phy(self, profile=None):
        """Get output phase of profile..
        :param profile: Profile phase requested (int)
        :returns: Output phase of DDS (float).
        """
        logging.debug("DdsTestDev.%r(%r)", inspect.stack()[0][3], profile)
        return self._phy

    def set_amp(self, value, profile=None):
        """Set amplitude tuning word of output signal on DDS.
        Take the input and output frequency as argument and set the adequat
        register in the DDS.
        :param value: Output amplitude value (int)
        :param profile: Profile to update between 0 to 7 (int)
        :returns: fsc register value if transfert is ok (int)
        """
        if self.is_connected() is False:
            logging.error("Device is not connected.")
            return None
        self._amp = value
        logging.debug("DdsTestDev.%r(%r, %r)",
                      inspect.stack()[0][3], value, profile)
        return value

    def get_amp(self, profile=None):
        """Get output amplitude tuning word of DDS.
        :param profile: Profile phase requested (int)
        :returns: Output amplitude tuning of DDS (float).
        """
        if self.is_connected() is False:
            logging.error("Device is not connected.")
            return None
        logging.debug("DdsTestDev.%r(%r)", inspect.stack()[0][3], profile)
        return self._amp

    def set_output_state(self, state=False):
        """Set output state.
        :param state: - False  Disable output. (bool)
                      - True   Enable CMOS output.
        :returns: None
        """
        raise NotImplemented

    def get_output_state(self):
        """Get output state.
        :returns: Output state (bool)
        """
        logging.debug("DdsTestDev.%r()", inspect.stack()[0][3])
        return True

    def set_pll_state(self, state=False):
        """Set PLL state.
        Note: A modification of the PLL state modify the output frequency.
        :param state: - False  Disable PLL. (bool)
                      - True   Enable PLL.
        :returns: None
        """
        logging.debug("DdsTestDev.%r(%r)", inspect.stack()[0][3], state)

    def get_pll_state(self):
        """Get PLL state.
        :returns: PLL state (bool)
        """
        logging.debug("DdsTestDev.%r()", inspect.stack()[0][3])
        return False

    def is_pll_locked(self):
        """Return the internal PLL lock (to the REF CLK input signal) state.
        :returns: True if the internal PLL is locked else return False (bool)
        """
        logging.debug("DdsTestDev.%r()", inspect.stack()[0][3])
        return False

    def set_pll_divider_factor(self, value):
        """Set PLL feedback divider value.
        :param factor: factor of PLL divider (between 20 to 510) (int)
        :returns: None
        """
        if self.is_connected() is False:
            logging.error("Device is not connected.")
        logging.debug("DdsTestDev.%r(%r)", inspect.stack()[0][3], value)

    def get_pll_divider_factor(self):
        """Get SysClk PLL divider factor.
        Note that here we get the overall divider factor, so the prescaler
        divider by 2 in the SysClk PLL divider block is include in the
        returned factor.
        :returns: factor of PLL divider (between 4 to 66) (int)
        """
        raise NotImplementedError

    def set_cp_current(self, value=0):
        """Set charge pump current value.
        :param value: charge pump current: - 0: 250 uA
                                           - 1: 375 uA
                                           - 2: off
                                           - 3: 125 uA
        :returns: None
        """
        raise NotImplementedError

    def get_cp_current(self):
        """Get charge pump current configuration value.
        Charge pump current: - 0: 250 uA
                             - 1: 375 uA
                             - 2: off
                             - 3: 125 uA
        :returns: charge pump current (int)
        """
        raise NotImplementedError

    def vco_calibration(self, value=None):
        logging.debug("DdsTestDev.%r(%r)", inspect.stack()[0][3], value)



# -*- coding: utf-8 -*-

"""package ad9xdds
author    Benoit Dubois
copyright FEMTO ENGINEERING, 2020
license   GPL v3.0+
brief     API to control AD9915 DDS development board from BBB.
details   AD9915 development board is configurable from Beagle Bone Black.
          Class allows configuration of principle DDS parameters (frequency,
          phase, amplitude, PLL handling, output handling...).

Beagle Bone Black cable connection:
    BBB                           AD9915 Dev Board
    P9_1 | P9_2     <-------------->  GND
    P9_17 (GPIO_4)  <-------------->  MPI00 (CSB)
    P9_22 (GPIO_2)  <-------------->  MPI01 (SCLK)
    P9_18 (GPIO_5)  <-------------->  MPI02 (SDIO)
    P9_21 (GPIO_3)  <-------------->  MPI03 (SDO)
    P9_15 (GPIO_48) <-------------->  IO_UPDATE (GPLIO0)
    P9_16 (GPIO_51) <-------------->  RESET_BUFF (GPLIO1)

    Profile selection (currently not implemented)
    PS0-BUF <---> GND
    PS1-BUF <---> GND
    PS2-BUF <---> GND

Configuration of AD9915 Dev Board jumpers:
    Disable USB communication
    P203 <---> Vcc (Disable)
    P204 <---> Vcc (Disable)
    P205 <---> Vcc (Disable)

    Serial programming mode (Datasheet AD9915 p.28):
    IOCFG3 <---> GND
    IOCFG2 <---> GND
    IOCFG1 <---> GND
    IOCFG0 <---> Vdd

    Others jumpers configuration
    SYNC_IO_BUFF <---> GND  (Disable I/O reset)
    EXTPDCTL-BUF <---> GND
    RESET-BUF    <---> OPEN
    DROVR-BUF    <---> GND
    DRHOLD-BUF   <---> GND
    DRCTL-BUF    <---> GND

"""

import logging
import time
from array import array
import usb.core
from binascii import hexlify
from Adafruit_BBIO.SPI import SPI
import Adafruit_BBIO.GPIO as GPIO


DEBUG = True

SPI_CLOCK_FREQUENCY = 100000
SPI_MODE = 0

RESET_PIN = "P9_16"
IOUPDATE_PIN = "P9_15"

IFMAX = 2500000000    # Input maximum frequency (Hz)
FTW_SIZE = 32         # Frequency Tuning Word register size (bit)
PHASE_SIZE = 16       # Phase register size (bit)
DAC_OUT_SIZE = 12     # Output DAC resolution (bit)
AMAX = (1 << DAC_OUT_SIZE) - 1  # Output maximum amplitude (a.u.)

regname_2_addr = {'CFR1': 0x0, 'CFR2': 0x1, 'CFR3': 0x2, 'CFR4': 0x3,
                  'DigRampLowerLimit': 0x4, 'DigRampUpperLimit': 0x5,
                  'RisingDigRampStepSize': 0x6, 'FallingDigRampStepSize': 0x7,
                  'DigRampRate': 0x8,
                  'LowerFreqJump': 0x9, 'UpperFreqJump': 0xA,
                  'P0Ftw': 0xB, 'P0PhaseAmp': 0xC,
                  'P1Ftw': 0xD, 'P1PhaseAmp': 0xE,
                  'P2Ftw': 0xF, 'P2PhaseAmp': 0x10,
                  'P3Ftw': 0x11, 'P3PhaseAmp': 0x12,
                  'P4Ftw': 0x13, 'P4PhaseAmp': 0x14,
                  'P5Ftw': 0x15, 'P5PhaseAmp': 0x16,
                  'P6Ftw': 0x17, 'P6PhaseAmp': 0x18,
                  'P7Ftw': 0x19, 'P7PhaseAmp': 0x1A,
                  'USR0': 0x1B}


class Ad9915DevBbb():
    """Class representing AD9915 development board through BeagleBoneBlack
    board. Currently class use profile mode P0 only
    """

    def __init__(self, ifreq=IFMAX):
        """The constructor.
        :param ifreq: Current input frequency in Hz (float))
        :returns: None
        """
        # Init properties related to FTDI device
        self._spi = SPI(0, 0)  # SPI bus
        self._init_spi()
        self._init_gpio()
        # Init class properties
        self._ifreq = ifreq
        self._profile = 0
        self._ofreq = None
        self._phy = None
        self._amp = None
        self._pll_state = None
        self._pll_doubler_state = None
        self._pll_factor = None

    def __del__(self):
        GPIO.cleanup()
        self._spi.close()

    def _init_spi(self):
        self._spi.msh = SPI_CLOCK_FREQUENCY
        self._spi.mode = SPI_MODE
        self._spi.bpw = 8  # Bits per word
        self._spi.threewire = False
        self._spi.cshigh = False
        self._spi.lsbfirst = False

    def _init_gpio(self):
        GPIO.setup(channel=RESET_PIN, direction=GPIO.OUT, initial=GPIO.LOW)
        GPIO.setup(channel=IOUPDATE_PIN, direction=GPIO.OUT, initial=GPIO.LOW)

    def connect(self):
        # Configure/Init DDS device
        ## Reset: needed at power-up (see datasheet p.40)
        self._master_reset()
        ## Default + SDIO input only
        self.write_w16(regname_2_addr['CFR1'], [0x00, 0x01, 0x00, 0x0A])
        ## Default Default + Enable profile mode
        self.write_w16(regname_2_addr['CFR2'], [0x00, 0x80, 0x09, 0x00])
        ## DAC cal: needed at power-up (see datasheet p.40)
        self.dac_calibration()
        self._init_properties()

    def _init_properties(self):
        """Init properties
        """
        # !warning! parameter order initialisation is important
        '''
        if self._pll_state is None:
            self._pll_state = self.get_pll_state()
        else:
            self.set_pll_state(self._pll_state)
        if self._pll_doubler_state is None:
            self._pll_doubler_state = self.get_pll_doubler_state()
        else:
            self.set_pll_doubler_state(self._pll_doubler_state)
        if self._pll_factor is None:
            self._pll_factor = self.get_pll_multiplier_factor()
        else:
            self.set_pll_multiplier_factor(self._pll_factor)
        '''
        if self._ofreq is None:
            self._ofreq = self.get_ofreq()
        else:
            self.set_ofreq(self._ofreq)
        if self._phy is None:
            self._phy = self.get_phy()
        else:
            self.set_phy(self._phy)
        if self._amp is None:
            self._amp = self.get_amp()
        else:
            self.set_amp(self._amp)

    def is_connected(self):
        """Return True if interface to DDS board is ready else return False.
        """
        '''if isinstance(self._spi, spi.SpiPort):
            return True
        return False
        '''
        return True

    def read(self, address, length=4):
        """
        :param register: register address(int)
        :param length: size of register in byte (int)
        :returns: register value (int)
        """
        msg = [0x80 + address]
        self._spi.writebytes(msg)
        self._io_update()
        reg_value = self._spi.readbytes(length)
        retval = 0
        for i, x in enumerate(reg_value):
            retval += x << (8*(len(reg_value)-i-1))
        return retval

    def write(self, address, value):
        """
        :param address: (int)
        :param value: (int)
        :returns: None
        """
        msg = [address]
        for i in range(3, -1, -1):
            msg.append((value >> (i * 8)) & 0xff)
        self._spi.writebytes(msg)
        self._io_update()
        self._spi.writebytes([])

    def read_w16(self, address, length=4):
        """
        :param register: register address(int)
        :param length: size of register in byte (int)
        :returns: register value in 16 bits word size (list of int)
        """
        msg = [0x80 + address]
        self._spi.writebytes(msg)
        self._io_update()
        retval = self._spi.readbytes(length)
        retval = [x for x in retval]
        return retval

    def write_w16(self, address, values):
        """
        :param address: (int)
        :param value: (list of int)
        :returns: None
        """
        msg = [address]
        for value in values:
            msg.append(value)
        self._spi.writebytes(msg)
        self._io_update()
        '''
        for value in msg:
            
            print("write_16.value:", value)
            
            self._spi.xfer2([value])
            self._io_update()
        '''
        #self._spi.writebytes([])

    def dac_calibration(self):
        """DAC calibration, needed after each power-up and every time REF CLK or
        the internal system clock is changed.
        :returns: None
        """
        # Calibration 1/2: enable
        self.write_w16(regname_2_addr['CFR4'], [0x01,0x05,0x21,0x20])
        # Calibration 2/2: disable
        self.write_w16(regname_2_addr['CFR4'], [0x00,0x05,0x21,0x20])

    def set_profile(self, profile):
        """Set profile currently in use. Curently not implemented.
        :param profile: Select profile in use between 0 to 7 (int)
        :returns: None
        """
        if profile not in range(0, 8):
            raise ValueError("Profile must be in range 0 to 7, here: %r",
                             profile)
        ## self._profile = profile

    def get_profile(self):
        """Get profile currently in use. Curently not implemented.
        :returns: profile currently in use (int)
        """
        return self._profile

    def get_sysfreq(self):
        """Get system frequency.
        Currently, does not support PLL.
        :returns: Current system frequency value (float)
        """
        '''if self._pll_state is True:
            if self._pll_doubler_state is True:
                doubler = 2.0
            else:
                doubler = 1.0
            factor = self._pll_factor
            sfreq = self._ifreq * doubler * factor
        else:
            sfreq = self._ifreq
        return sfreq
        '''
        return self._ifreq

    def set_ifreq(self, value):
        """Set input frequency.
        :param value: Input frequency value (float)
        :returns: None
        """
        self._ifreq = value
        self.dac_calibration()  # Needed!
        logging.debug("Set input frequency: %r", value)
        # Update DDS device because ofreq = f(ifreq)
        self.set_ofreq(self._ofreq)

    def get_ifreq(self):
        """Get input frequency.
        :returns: Current input frequency value (float)
        """
        return self._ifreq

    def set_ofreq(self, value, profile=None):
        """Set output frequency to current DDS profile if profile parameter is
        None or set output frequency of requested DDS profile.
        Return the actual output frequency (see _actual_ofreq() method).
        :param value: Output frequency value (float).
        :param profile: Profile to update between 0 to 7 (int)
        :returns: Actual output frequency (float)
        """
        if self.is_connected() is False:
            logging.error("Device is not connected.")
            return None
        if profile not in range(0, 8) and profile is not None:
            raise ValueError("Profile not in range 0 to 7, here: %r", profile)
        if profile is None:
            profile = 0
        regname = 'P{:d}Ftw'.format(profile)
        # Compute the Frequency Tuning Word (FTW).
        ftw = int((value * (1 << FTW_SIZE)) / self.get_sysfreq())
        
        print("set_ofreq.ftw", ftw)
        
        # Send data to device
        self.write(regname_2_addr[regname], ftw)
        # Return the actual output frequency.
        self._ofreq = self._actual_ofreq(self.get_sysfreq(), ftw, FTW_SIZE)
        logging.debug("Set output frequency: %r", self._ofreq)
        return self._ofreq

    def get_ofreq(self, profile=None):
        """Get output frequency of current DDS profile if profile parameter is
        None or return output frequency of requested DDS profile.
        :param profile: Profile output frequency requested (int)
        :returns: Output frequency of DDS profile (float).
        """
        if self.is_connected() is False:
            logging.error("Device is not connected.")
            return None
        if profile not in range(0, 8) and profile is not None:
            raise ValueError("Profile not in range 0 to 7, here: %r", profile)
        if profile is None:
            profile = 0
        regname = 'P{:d}Ftw'.format(profile)
        ftw = self.read(regname_2_addr[regname])

        print("get_ofreq.ftw", ftw)
        
        self._ofreq = self._actual_ofreq(self.get_sysfreq(), ftw, FTW_SIZE)
        return self._ofreq

    def set_phy(self, value, profile=None):
        """Set phase of output signal on DDS.
        Take the queried output phase (in degree) as argument and set
        the adequat register in the DDS.
        :param value: Output phase value (float).
        :param profile: Profile to update between 0 to 7 (int)
        :returns: Actual output phase (float)
        """
        if self.is_connected() is False:
            logging.error("Device is not connected.")
            return None
        if profile not in range(0, 8) and profile is not None:
            raise ValueError("Profile not in range 0 to 7, here: %r", profile)
        if profile is None:
            profile = 0
        regname = 'P{:d}PhaseAmp'.format(profile)
        # Phase and amplitude output is handled by a common register.
        # Before writing new phase, we need to get current amplitude
        # and take care to not reset its value.
        reg_value = self.read_w16(regname_2_addr[regname])
        asf_list = reg_value[2:]
        phy = int((value * (1 << PHASE_SIZE)) / 360)
        phy_list = self._int_2_byte_list(phy, 2)
        msg = asf_list + phy_list
        self.write_w16(regname_2_addr[regname], msg)
        phy = self._actual_phy(phy, PHASE_SIZE)  # Return the actual phase
        logging.debug("Set phase: %r", phy)
        return phy

    def get_phy(self, profile=None):
        """Get output phase of profile..
        :param profile: Profile phase requested (int)
        :returns: Output phase of DDS (float).
        """
        if self.is_connected() is False:
            logging.error("Device is not connected.")
            return None
        if profile not in range(0, 8) and profile is not None:
            raise ValueError("Profile not in range 0 to 7, here: %r", profile)
        if profile is None:
            profile = 0
        regname = 'P{:d}PhaseAmp'.format(profile)
        reg_value = self.read_w16(regname_2_addr[regname])
        # Extract phase value from register value.
        phy = (reg_value[1] << 8) + reg_value[0]
        return self._actual_phy(phy, PHASE_SIZE)  # return the actual phase.

    def set_amp(self, value, profile=None):
        """Set amplitude tuning word of output signal on DDS.
        Take the input and output frequency as argument and set the adequat
        register in the DDS.
        :param value: Output amplitude value (int)
        :param profile: Profile to update between 0 to 7 (int)
        :returns: fsc register value if transfert is ok (int)
        """
        if self.is_connected() is False:
            logging.error("Device is not connected.")
            return None
        if profile not in range(0, 8) and profile is not None:
            raise ValueError("Profile not in range 0 to 7, here: %r", profile)
        if profile is None:
            profile = 0
        # If value is out of range, bound value and raise Warning.
        if not(0 <= value <= AMAX):
            logging.warning("Amplitude value out of range: %r", value)
            value = self._bound_value(value, 0, AMAX)
        regname = 'P{:d}PhaseAmp'.format(profile)
        # Phase and amplitude output is handled by a common register.
        # Before writing new amplitude, we need to get current phase
        # and take care to not reset its value.
        reg_value = self.read_w16(regname_2_addr[regname])
        phy_list = reg_value[:2]
        asf_list = self._int_2_byte_list(value, 2)
        msg = asf_list + phy_list
        self.write_w16(regname_2_addr[regname], msg)
        logging.debug("Set output amplitude: %r", value)
        return value

    def get_amp(self, profile=None):
        """Get output amplitude tuning word of DDS.
        :param profile: Profile phase requested (int)
        :returns:  Output amplitude tuning of DDS (float).
        """
        if self.is_connected() is False:
            logging.error("Device is not connected.")
            return None
        if profile not in range(0, 8) and profile is not None:
            raise ValueError("Profile not in range 0 to 7, here: %r", profile)
        if profile is None:
            profile = 0
        regname = 'P{:d}PhaseAmp'.format(profile)
        reg_value = self.read_w16(regname_2_addr[regname])
        # Extract amplitude value from register value.
        asf = (reg_value[3] << 8) + reg_value[2]
        return asf

    def set_output_state(self, state=False):
        """Set output state.
        :param state: - False  Disable output. (bool)
                      - True   Enable CMOS output.
        :returns: None
        """
        raise NotImplemented

    def get_output_state(self):
        """Get output state.
        :returns: Output state (bool)
        """
        reg_value = self.read(regname_2_addr['CFR1'])
        output_state_bit = 16
        return self._check_bit_set(reg_value, output_state_bit)

    def set_pll_state(self, state=False):
        """Set PLL state.
        Note: A modification of the PLL state modify the output frequency.
        :param state: - False  Disable PLL. (bool)
                      - True   Enable PLL.
        :returns: None
        """
        reg_value = self.read(regname_2_addr['CFR3'])
        pll_state_bit = 2
        if state is True:
            reg_value |= 1 << pll_state_bit
        else:
            reg_value &= ~(1 << pll_state_bit)
        self.write(regname_2_addr['CFR3'], reg_value)

    def get_pll_state(self):
        """Get PLL state.
        :returns: PLL state (bool)
        """
        reg_value = self.read(regname_2_addr['CFR3'])
        pll_state_bit = 2
        return self._check_bit_set(reg_value, pll_state_bit)

    def is_pll_locked(self):
        """Return the internal PLL lock (to the REF CLK input signal) state.
        :returns: True if the internal PLL is locked else return False (bool)
        """
        reg_value = self.read(regname_2_addr['USR0'])
        pll_lock_bit = 24
        return self._check_bit_set(reg_value, pll_lock_bit)

    def set_pll_divider_factor(self, value):
        """Set PLL feedback divider value.
        :param factor: factor of PLL divider (between 20 to 510) (int)
        :returns: None
        """
        if self.is_connected() is False:
            logging.error("Device is not connected.")
            return None
        if not(20 <= value <= 510):
            logging.error("PLL divider factor value out of range: %r", factor)
            value = self._bound_value(value, 20 , 510)
        cfr3 = self.read_w16(regname_2_addr['CFR3'])
        cfr3[1] = value
        self.write_w16(regname_2_addr['CFR3'], cfr3)
        logging.debug("Set PLL divider factor: %r", value)
        return value

    def get_pll_divider_factor(self):
        """Get SysClk PLL divider factor.
        Note that here we get the overall divider factor, so the prescaler
        divider by 2 in the SysClk PLL divider block is include in the
        returned factor.
        :returns: factor of PLL divider (between 4 to 66) (int)
        """
        raise NotImplemented

    def set_cp_current(self, value=0):
        """Set charge pump current value.
        :param value: charge pump current: - 0: 250 uA
                                           - 1: 375 uA
                                           - 2: off
                                           - 3: 125 uA
        :returns: None
        """
        raise NotImplemented

    def get_cp_current(self):
        """Get charge pump current configuration value.
        Charge pump current: - 0: 250 uA
                             - 1: 375 uA
                             - 2: off
                             - 3: 125 uA
        :returns: charge pump current (int)
        """
        raise NotImplemented

    def vco_calibration(self, value=None):
        """VCO calibration process.
        :param value: vco range: - 0 = low range (int)
                                 - 1 = high range
                                 - others = autorange
        :returns: None
        """
        raise NotImplemented

    @staticmethod
    def _actual_ofreq(ifreq, ftw, rsize):
        """Return the actual output frequency.
        Due to the resolution of the DDS, the actual output frequency
        may be (a bit) different than the queried.
        :param ftw: Frequency Tuning Word register value (int).
        :param rsize: Size of register (int).
        :returns: Actual output frequency (float).
        """
        return (ftw * ifreq) / (1 << rsize)

    @staticmethod
    def _actual_phy(dphy, bit):
        """Return the actual output phase.
        Due to the resolution of the DDS, the actual output phase
        may be (a bit) different than the queried.
        :param dphy: Phase register value (int).
        :param bit: Number of bits used for the phase resolution (int)
        :returns: Actual output phase offset in degree (float).
        """
        return float(360 * dphy) / (1 << bit)

    @staticmethod
    def _int_2_byte_list(value, size):
        """Take an integer and split it in a list of the corresponding
        'size' 8 bits word (byte).
        For example:
            17179869 in hexadecimal = 0x10624dd
            0x10624dd splitted in byte = [0x01, 0x06, 0x24, 0xdd]
            [0x01, 0x06, 0x24, 0xdd] in base 10 = [1, 6, 36, 221]
            => _int_2_byte_list(17179869, 4) = [1, 6, 36, 221]
        :param value: an integer to split (int)
        ;param size: sier of output list (int)
        :returns: list of byte (list of int)
        """
        value_format = '0{:d}x'.format(size * 2)
        value = format(value, value_format)
        return [int(value[i:i+2], 16) for i in range(0, size*2, 2)]

    @staticmethod
    def _bound_value(value, vmin, vmax):
        """Check that a value is included in the range [min, max], if not the value
        is bounded to the range, ie:
        - if value < min  ->  min = value
        - if value > max  ->  max = value
        :param value: Value that is checked
        :param vmin: Minimum valid value.
        :param vmax: Maximum valid value.
        :returns: Bounded value.
        """
        if value < vmin:
            logging.warning("Parameter %s out of range (%f). Set to: %f",
                            val_name, value, vmin)
            return vmin
        elif value > vmax:
            logging.warning("Parameter %s out of range (%f). Set to: %f",
                            val_name, value, vmax)
            return vmax
        else:
            return value

    @staticmethod
    def _check_bit_set(word, n):
        """Check if n-th bit of word is set or not.
        :param word: word to check (int)
        :param n: n-th bit of work to check (int)
        :returns: True is n-th bit is set else False (bool)
        """
        if word & (1 << n):
            return True
        return False

    def _io_update(self):
        """Generate IO update event: transfer written data from I/O buffer to
        the coresponding internal registers.
        :returns: None
        """
        GPIO.output(IOUPDATE_PIN, GPIO.HIGH)
        #time.sleep(0.001)
        GPIO.output(IOUPDATE_PIN, GPIO.LOW)
        logging.debug("IO update")

    def _master_reset(self):
        """Master reset: clears all memory elements and sets registers to
        default values. Required after power up.
        :returns: None
        """
        GPIO.output(RESET_PIN, GPIO.HIGH)
        time.sleep(int(1 / (1000 * SPI_CLOCK_FREQUENCY)))
        GPIO.output(RESET_PIN, GPIO.LOW)
        logging.debug("Master reset")


# =============================================================================
if __name__ == '__main__':
    # Setup basic logger
    LOG_FORMAT = '%(asctime)s %(levelname)s %(filename)s (%(lineno)d): ' \
        + '%(message)s'
    logging.basicConfig(format=LOG_FORMAT, level=logging.DEBUG)

    pylogger = logging.getLogger('pyftdi')
    #pylogger.setLevel(logging.DEBUG)

    ##print("Interfaces:", Ad9915DevFtdi232HM().get_interfaces())

    OFREQ = 80000000.0
    #OFREQ = 38146.3905796
    PHY = 15
    AMP = 478
    DDS = Ad9915DevBbb()
    DDS.connect()
    #DDS.set_pll_divider_factor(250)
    #DDS.set_pll_state(True)
    #print("PLL lock:", DDS.is_pll_locked())
    #DDS._io_update()
    DDS.set_ifreq(2.5e9)
    print("CFR2:", DDS.read_w16(regname_2_addr['CFR2']))
    print(DDS.set_ofreq(OFREQ))
    print("Get output frequency")
    print("OFREQ:", DDS.get_ofreq())
    #print("Get amplitude")
    #print("AMP 1:", DDS.get_amp())
    #DDS.set_phy(PHY)
    #print("Get phase")
    #print("PHY:", DDS.get_phy())
    #DDS.set_amp(AMP)
    #print("Get amplitude")
    #print("AMP 2:", DDS.get_amp())



# -*- coding: utf-8 -*-

"""package ad9xdds
author    Benoit Dubois
copyright FEMTO ENGINEERING, 2019
license   GPL v3.0+
brief     API to control AD9852/4.
details   Ad9854Dev: AD9852/4 DDS development card class.
          Note: currently only support parallel interface connection.
"""

import time
import importlib
from dds_core import AbstractDds, bound_value, split_len


# =============================================================================
class Ad9854Dev(AbstractDds):
    """Class representing AD9854 development card.
    Board integrates a parallel port for transmission with PC. AD9854 device
    can be use with a serial or a parallel interface paradigm while parallel
    port of PC is being used for communication. Note that board parallel
    interface allow only register writing.
    """

    FTW_SIZE = 48               # Frequency Tuning Word size (bit)
    PHASE_SIZE = 14             # Phase register size (bit)
    DAC_OUT_SIZE = 12           # Output DAC resolution (bit)
    IFMAX = 300000000           # Input maximum frequency (Hz)
    OFMAX = 120000000           # Output maximum frequency (Hz)
    AMAX = (1 << DAC_OUT_SIZE) - 1  # Output maximum amplitude (a.u.)

    WR_BIT = 0
    RD_BIT = 1
    MRES_BIT = 2
    UDCLK_BIT = 3
    SPMOD_BIT = 4   # Serial/Parallel mode selection
    ORAMP_BIT = 5
    FDATA_BIT = 6

    _regName2addr = {'Phy1m': 0x0, 'Phy1l': 0x1, 'Phy2m': 0x2, 'Phy2l': 0x3,
                     'Ftw1_5': 0x4, 'Ftw1_4': 0x5, 'Ftw1_3': 0x6,
                     'Ftw1_2': 0x7, 'Ftw1_1': 0x8, 'Ftw1_0': 0x9,
                     'Ftw2_5': 0xa, 'Ftw2_4': 0xb, 'Ftw2_3': 0xc,
                     'Ftw2_2': 0xd, 'Ftw2_1': 0xe, 'Ftw2_0': 0xf,
                     'DFw_5': 0x10, 'DFw_4': 0x11, 'DFw_3': 0x12,
                     'DFw_2': 0x13, 'DFw_1': 0x14, 'DFw_0': 0x15,
                     'UpClk_3': 0x16, 'UpClk_2': 0x17, 'UpClk_1': 0x18,
                     'UpClk_0': 0x19, 'RrClk_2': 0x1a, 'RrClk_1': 0x1b,
                     'RrClk_0': 0x1c, 'Ctr_3': 0x1d, 'Ctr_2': 0x1e,
                     'Ctr_1': 0x1f, 'Ctr_0': 0x20, 'OskIm': 0x21,
                     'OskIl': 0x22, 'OskQm': 0x23, 'OskQl': 0x24,
                     'OskRr': 0x25, 'Qdacm': 0x26, 'Qdacl': 0x27}

    def __init__(self, port=0):
        """The constructor."""
        # Import specific parallel device packages
        try:
            parallel = importlib.import_module("parallel")
        except KeyError as ex:
            print("Problem during parallel class import.")
            print("DDS class %s not found" % str(ex))
        except Exception as ex:
            print("Error during parallel class import: %s" % str(ex))
        try:
            self.pdev = parallel.Parallel(port)
        except OSError:
            print("Error during parallel port initialization.")
            print("Is lp module unloaded and ppdev module loaded?")
            print("Have you write permissions to parallel port file?")
            raise
        except IOError:
            print("Error during parallel port initialization.")
            print("Parallel device file %s is not found" % str(port))
            raise
        except Exception:
            print("Error during parallel port initialization.")
            raise
        # Reset DDS device
        self.reset_dev()
        # Initialize parallel port Ctrl pins
        self.pdev.setDataStrobe(1)
        self.pdev.setAutoFeed(1)
        self.pdev.setInitOut(0)  # Note: enable ADDR0 line
        self.pdev.setSelect(1)
        # Initialize data lines
        self.pdev.setData(0xff)
        self._latch_data()
        # Initialize address lines
        self.pdev.setData(0xff)
        self._latch_addr()
        # Initialize WRBAR,RDBAR,PMODE high.
        # Initialize RESET,UDCLK low.
        # Initialize FDATA,ORAMP accordingly.
        # Note: PMODE high enables ADDR1 line.
        self.pdev.setData(0xff - ((1 << self.WR_BIT) +
                                  (1 << self.RD_BIT) +
                                  (1 << self.SPMOD_BIT)))
        self._latch_ctrl()
        # Initialize control register.
        # Ctr_3 disable power_down mode
        self.set_reg(self._regName2addr['Ctr_3'], 0x00)
        # Disable PLL by default
        self.set_reg(self._regName2addr['Ctr_2'], 0x64)
        # Ctr_1 must be configured because its SPMOD_BIT controls
        # the comportment of the 'IO UD CLK' pin.
        self.set_reg(self._regName2addr['Ctr_1'], 0x00)
        # Set 'OSK EN' high to allow user control of amplitude
        self.set_reg(self._regName2addr['Ctr_0'], 0x20)

    def set_ifreq(self, ifreq):
        """Set input frequency.
        :param ifreq: Input frequency value (float)
        :returns: None
        """
        self._ifreq = float(ifreq)

    def get_ifreq(self):
        """Get input frequency.
        :returns: Current input frequency value (float)
        """
        return self._ifreq

    def reset_dev(self):
        """ Reset DDS device: set MRESET pin @ 1 during 10 system clock cycles.
        """
        self.pdev.setData(0xff - ((1 << self.WR_BIT) +
                                  (1 << self.RD_BIT) +
                                  (1 << self.MRES_BIT) +
                                  (1 << self.SPMOD_BIT)))
        self._latch_ctrl()
        time.sleep(0.1)  # Conservative value for 10 system clock cycles
        self.pdev.setData(0xff - ((1 << self.WR_BIT) +
                                  (1 << self.RD_BIT) +
                                  (1 << self.SPMOD_BIT)))
        self._latch_ctrl()

    def _latch_data(self):
        """ Macro for latching parallel data on data bus of the device.
        """
        self.pdev.setDataStrobe(0)
        self.pdev.setDataStrobe(1)

    def _latch_addr(self):
        """ Macro for latching parallel data on address bus of the device.
        """
        self.pdev.setAutoFeed(0)
        self.pdev.setAutoFeed(1)

    def _latch_ctrl(self):
        """ Macro for latching bits of control of device (\WR, \RD, RESET...).
        """
        self.pdev.setSelect(0)
        self.pdev.setSelect(1)

    def _upclock(self):
        """ Update clock (external update mode).
        """
        # Set WRBAR,RDBAR,UDCLK,PMODE high.
        # Set RESET low.
        # Set FDATA,ORAMP accordingly.
        self.pdev.setData(0xff - ((1 << self.WR_BIT) +
                                  (1 << self.RD_BIT) +
                                  (1 << self.UDCLK_BIT) +
                                  (1 << self.SPMOD_BIT)))
        self._latch_ctrl()
        # Set WRBAR,RDBAR,PMODE high.
        # Set RESET,UDCLK low.
        # Set FDATA,ORAMP accordingly.
        self.pdev.setData(0xff - ((1 << self.WR_BIT) +
                                  (1 << self.RD_BIT) +
                                  (1 << self.SPMOD_BIT)))
        self._latch_ctrl()

    def _actual_phy(self, dphy):
        """Public function for specific AD9854 product.
        arg    dphy Phase register value (int).
        return      Actual output phase offset in degree (float).
        """
        self._actual_phy(dphy, self.PHASE_SIZE)

    def set_reg(self, address, value):
        """Write value @ address register.
        arg    address  Address of the register to write (int).
        arg    value    Value to write (int).
        return          None
        Rmp: ici l'ecriture dans les registres fonctionnent si on considère
        les registres en mode "parallèle", alors que la lecture utilise le
        mode série. => passer en mode série également
        """
        # Prepare data and address bus.
        self.pdev.setData(0xff-value)  # Complement binary code
        self._latch_data()
        self.pdev.setData(0xff-address)  # Complement binary code
        self._latch_addr()
        # Write value @ address in the I/O buffer register.
        # WRBAR,RESET,UDCLK = lo
        # RDBAR,PMODE = hi
        self.pdev.setData(0xff - ((1 << self.RD_BIT) +
                                  (1 << self.SPMOD_BIT)))
        self._latch_ctrl()
        # RESET,UDCLK = lo
        # WRBAR,RDBAR,PMODE = hi
        self.pdev.setData(0xff - ((1 << self.WR_BIT) +
                                  (1 << self.RD_BIT) +
                                  (1 << self.SPMOD_BIT)))
        self._latch_ctrl()

    def get_reg(self, address):
        """Read value @ address.
        arg    address  Address of the register to read (int)
        return   Register value (int)
        """
        self.pdev.setData(0xff - 1)
        self._latch_ctrl()
        self.pdev.setData(0xff - 0)
        self._latch_ctrl()
        # Build serial read request message:
        # Read (1b) | Don't Care (3b) | address (4b)
        msg = "1000" + format(address, '04b')
        prv_str = "0"
        # Get eight bits.
        for i in range(8):
            tmp_str = msg[i]
            # Change data bit only if necessary.
            if ((tmp_str == "1") and (prv_str == "0")):
                self.pdev.setData(0xfe)
                self._latch_addr()
                prv_str = "1"
            elif ((tmp_str == "0") and (prv_str == "1")):
                self.pdev.setData(0xff)
                self._latch_addr()
                prv_str = "0"
            # Set SCLK high.
            self.pdev.setData(0xff - 1)
            self._latch_ctrl()
            if (i < 8):
                # Set SCLK low.
                self.pdev.setData(0xff - 0)
                self._latch_ctrl()
        self.pdev.setData(0xff)
        self._latch_addr()
        # Set line for 2 wire mode (default mode).
        self.pdev.setInitOut(1)
        # Default: MSB first.
        bits_to_read = address
        # Determine number of bits to read.
        if bits_to_read == 10:
            bits = 8
        elif (bits_to_read == 0 or bits_to_read == 1 or bits_to_read == 8 or
              bits_to_read == 9 or bits_to_read == 11):
            bits = 16
        elif bits_to_read == 6:
            bits = 24
        elif (bits_to_read == 5 or bits_to_read == 7):
            bits = 32
        elif (bits_to_read == 2 or bits_to_read == 3 or bits_to_read == 4):
            bits = 48
        else:
            raise ValueError("Bad address to read")
        # Initialize readback string.
        tmp_str = ""
        for i in range(bits):
            # Take SCLK low.
            self.pdev.setData(0xff - 0)
            self._latch_ctrl()
            # Readback port and mask unwanted bits.
            res = self.pdev.getInError()
            # Build string.
            # Note: Data are inverted on customer eval board.
            if res is True:
                tmp_str = tmp_str + "0"
            else:
                tmp_str = tmp_str + "1"
            # Take SCLK high.
            self.pdev.setData(0xff - 1)
            self._latch_ctrl()
        deta = tmp_str
        # De-Set line for 2 wire mode (default mode).
        # if (2WireModeEnabled):
        self.pdev.setInitOut(0)
        self.pdev.setData(0xff - 3)
        self._latch_ctrl()
        return int(deta, 2)

    def set_ofreq(self, ofreq):
        """Set output frequency on DDS.
        Take the input and output frequency as argument and set the adequat
        register in the DDS and return the actual output frequency (see
        _actual_ofreq()).
        output frequency may be (a bit) different than the requested frequency.
        arg    ofreq  Output frequency (float).
        return        Actual output frequency (float)
        """
        # Compute the Frequency Tuning Word (FTW).
        ftw = int((ofreq * (1 << self.FTW_SIZE)) / self.ifreq)
        # Prepare list of value to send to the 6 FTW registers.
        ftw_val_list = split_len(format(ftw, '012x'), 2)
        # Send values to the 6 FTW0 registers of the DDS.
        # TODO Search a better way to find reg address.
        for idx, ftw_val in enumerate(ftw_val_list):
            try:
                self.set_reg(self._regName2addr['Ftw1_5']+idx,
                             int(ftw_val, 16))
            except Exception as ex:
                print("Setting frequency failed:", str(ex))
                raise
        # Request clock update (manual mode)
        self._upclock()
        # Return the actual output frequency.
        return self._actual_ofreq(self.get_sysfreq(), ftw, self.FTW_SIZE)

    def get_ofreq(self):
        """Get output frequency of DDS.
        return:       Output frequency of DDS (float).
        """
        # Get back the Frequency Tuning Word (FTW) from device.
        ftw = 0
        for idx in range(0, 6):
            ftw += self.get_reg(self._regName2addr['Ftw1_5']+idx) << (idx*8)
        return self._actual_ofreq(self.get_sysfreq(), ftw, self.FTW_SIZE)

    def set_phy(self, phy):
        """Set phase of output signal on DDS.
        Take the queried output phase (in degree) as argument and set
        the adequat register in the DDS.
        arg    phy  Output phase (float).
        return      Actual output phase (float)
        """
        # Raise error if value is out of range.
        if not(0 <= phy < 360):
            raise ValueError('Phase value out of range.')
        # Compute dphy (Delta_phase).
        dphy = int((phy * (1 << 14)) / 360)
        # Prepare list of value to send to the dphy registers.
        dphy_val_list = split_len(format(dphy, '04x'), 2)
        # Send values to the dphy registers of the DDS.
        # TODO Search a better way to find reg address.
        for idx, val in enumerate(dphy_val_list):
            self.set_reg(self._regName2addr['Phy1m']+idx, int(val, 16))
        # Request clock update (manual mode)
        self._upclock()
        # Return the actual output phase
        return self._actual_phy(dphy)

    def get_phy(self):
        """Get output phase of DDS.
        return:  Output phase of DDS (float).
        """
        # Get back the dphy register content
        dphy = 0
        for idx in range(0, 2):
            dphy += self.get_reg(self._regName2addr['Phy1m']+idx) << (idx*8)
        # return the ofreq.
        return self._actual_phy(dphy)

    def set_amp(self, amp):
        """Set amplitude tuning word of output signal on DDS.
        Take the input and output frequency as argument and set the adequat
        register in the DDS.
        :param amp: Output amplitude (int).
        :returns: Amplitude register value if transfert is ok (int).
        """
        # If value is out of range, bound value and raise Warning.
        if not(0 <= amp <= self.AMAX):
            amp = bound_value(amp, 0, self.AMAX)
            raise ValueError('Amplitude tuning value out of range.')
        # Prepare list of value to send to the amp registers.
        amp_val_list = split_len(format(amp, '04x'), 2)
        # Send values to the amp registers of the DDS.
        # TODO Search a better way to find reg address.
        for idx, val in enumerate(amp_val_list):
            self.set_reg(self._regName2addr['OskIm']+idx, int(val, 16))
        # Request clock update (manual mode)
        self._upclock()
        # Return the amplitude value if transfert is ok.
        return amp

    def get_amp(self):
        """Get output amplitude tuning word of DDS.
        return:  Output amplitude tuning of DDS (float).
        """
        # Get back the AMP register (Full Scale) register content
        amp = 0
        for idx in range(0, 2):
            amp += self.get_reg(self._regName2addr['OskIm']+idx) << (idx*8)
        # Return the amplitude tuning word.
        return amp


# =============================================================================
if __name__ == '__main__':
    IFREQ = 1000000.0
    OFREQ = 100000.0
    DDS = Ad9854Dev()
    DDS.set_ifreq(IFREQ)
    DDS.set_ofreq(OFREQ)



# -*- coding: utf-8 -*-

"""package ad9xdds
author    Benoit Dubois
copyright FEMTO ENGINEERING, 2018-2025
license   GPL v3.0+
brief     API to control AD9912 DDS development board.
details   AD9912 development board is configurable by USB through an USB
          to serial transceiver from Cypress (FX2 device). Class allows
          configuration of principle DDS parameters (frequency, phase,
          amplitude, PLL handling, output handling...).
"""

import logging
from array import array
import usb.core
from iopy.fx2 import Fx2Core


AD9912DEV_PRODUCT_ID_LIST = [0xee09, 0xee24]  # Dev board has plural product_id
VENDOR_ID = 0x0456
PRODUCT_ID = AD9912DEV_PRODUCT_ID_LIST[0]

FTW_SIZE = 48               # Frequency Tuning Word register size (bit)
PHASE_SIZE = 14             # Phase register size (bit)
DAC_OUT_SIZE = 12           # Output DAC resolution (bit)

IFMAX = 1000000000.0        # Input maximum frequency (Hz)
OFMAX = 400000000.0         # Output maximum frequency (Hz)
AMAX = (1 << DAC_OUT_SIZE) - 1  # Output maximum amplitude (a.u.)


# =============================================================================
def split_len(seq, length):
    """Take data of sequence type and return a list containing the n-sized
    pieces of the string.
    For example: split_len('aabbccdd', 2) => ['aa', 'bb', 'cc', 'dd']
    arg    seq     Any sequences that support slicing and len().
    arg    length  Size of slice (int).
    return         Splited sequence (list of str).
    """
    return [seq[i:i+length] for i in range(0, len(seq), length)]


# =============================================================================
def bound_value(value, vmin, vmax, val_name=""):
    """Check that a value is included in the range [min, max], if not the value
    is bounded to the range, ie:
    - if value < min  ->  min = value
    - if value > max  ->  max = value
    :param value: Value that is checked
    :param vmin: Minimum valid value.
    :param vmax: Maximum valid value.
    :returns: Bounded value.
    """
    if value < vmin:
        logging.warning("Parameter %s out of range (%f). Set to: %f",
                        val_name, value, vmin)
        return vmin
    if value > vmax:
        logging.warning("Parameter %s out of range (%f). Set to: %f",
                        val_name, value, vmax)
        return vmax
    return value


# =============================================================================
class Ad9912Dev():
    """Class representing AD9912 development card.
    Board integrates an USB port for transmission with PC and a Cypress FX2
    device is used as USB transceiver.
    """

    _CONFIG_ID     = 1
    _INTERFACE_ID  = 0  # The interface we use to talk to the device
    _ALT_INTF_ID   = 0  # The alternate interface use to talk to the device
    _IN_EP_ID      = 1  # EP1_IN @0x81 = 129 -> small endpoint
    _OUT_EP_ID     = 0  # EP1_OUT @0x01 = 1  -> small endpoint
    _IN_DAT_EP_ID  = 5  # EP8 @0x88 = 136    -> large endpoint
    _OUT_DAT_EP_ID = 3  # EP4 @0x04 = 4      -> large endpoint

    _regName2addr = {'SerPortCfg': 0x0, 'PartIdLsb': 0x2, 'PartIdMsb': 0x3,
                     'SerOpt': 0x4, 'SerOptClr': 0x5, 'PDEn': 0x10,
                     'ResetAuto': 0x12, 'ResetNAuto': 0x13, 'Ndivider': 0x20,
                     'PllParam': 0x22, 'SdivLsb': 0x104, 'SdivMsb': 0x105,
                     'SdivCfg': 0x106, 'Ftw0_0': 0x1a6, 'Ftw0_1': 0x1a7,
                     'Ftw0_2': 0x1a8, 'Ftw0_3': 0x1a9, 'Ftw0_4': 0x1aa,
                     'Ftw0_5': 0x1ab, 'ACphase': 0x1ac, 'ADphase': 0x1ad,
                     'HstlDrv': 0x200, 'CmosDrv': 0x201, 'DacFsLsb': 0x40b,
                     'DacFsMsb': 0x40c, 'SpurA': 0x500, 'SpurAmag': 0x501,
                     'SpurAphLsb': 0x503, 'SpurAphMsb': 0x504,
                     'SpurB': 0x505, 'SpurBmag': 0x506,
                     'SpurBphLsb': 0x508, 'SpurBphMsb': 0x509}

    def __init__(self, ifreq=IFMAX):
        """The constructor.
        :param ifreq: Current input frequency in Hz (float)
        :returns: None
        """
        self._fx2 = Fx2Core()  # USB interface
        self._ifreq = float(ifreq)
        self._ofreq = None
        self._phy = None
        self._amp = None
        self._pll_state = None
        self._pll_doubler_state = None
        self._pll_factor = None

    def connect(self, vendor_id=VENDOR_ID, product_id=PRODUCT_ID,
                bus=None, address=None):
        """Configure USB device defined by is VID:PID.
        :param vendor_id: identification number of vendor (int)
        :param product_id: identification number of product (int)
        :param bus: bus attribute of device (int)
        :param address: address attribute of device (int)
        :returns: True if connection is Ok else returns False (bool)
        """
        if self._fx2.connect(vendor_id, product_id, bus, address) is False:
            return False
        try:
            self._fx2.config(config_id=self._CONFIG_ID,
                             interface_id=self._INTERFACE_ID,
                             alt_interface_id=self._ALT_INTF_ID,
                             in_ep_id=self._IN_EP_ID,
                             out_ep_id=self._OUT_EP_ID,
                             in_dat_ep_id=self._IN_DAT_EP_ID,
                             out_dat_ep_id=self._OUT_DAT_EP_ID)
        except Exception as ex:
            logging.error("FX2 USB config failed: %r", ex)
            return False
        self.set_led(False)  # Disable LED blinking of development board
        self._init()
        logging.debug("Connected to AD9912 board")
        return True

    def disconnect(self):
        self._fx2.disconnect()

    def is_connected(self):
        return self._fx2.is_connected()

    def _init(self):
        """Init properties
        """
        # !warning! parameter order initialisation is important
        if self._pll_state is None:
            self._pll_state = self.get_pll_state()
        else:
            self.set_pll_state(self._pll_state)
        if self._pll_doubler_state is None:
            self._pll_doubler_state = self.get_pll_doubler_state()
        else:
            self.set_pll_doubler_state(self._pll_doubler_state)
        if self._pll_factor is None:
            self._pll_factor = self.get_pll_multiplier_factor()
        else:
            self.set_pll_multiplier_factor(self._pll_factor)
        if self._ofreq is None:
            self._ofreq = self.get_ofreq()
        else:
            self.set_ofreq(self._ofreq)
        if self._phy is None:
            self._phy = self.get_phy()
        else:
            self.set_phy(self._phy)
        if self._amp is None:
            self._amp = self.get_amp()
        else:
            self.set_amp(self._amp)

    def set_ifreq(self, value):
        """Set input frequency.
        :param value: Input frequency value (float)
        :returns: ifreq value (float).
        """
        self._ifreq = float(value)
        logging.debug("Set input frequency: %r", value)
        # Update DDS device because ofreq = f(ifreq)
        self.set_ofreq(self._ofreq)
        return value

    def get_ifreq(self):
        """Get input frequency.
        :returns: Current input frequency value (float)
        """
        return self._ifreq

    def get_sysfreq(self):
        """Get system frequency.
        :returns: Current system frequency value (float)
        """
        if self._pll_state is True:
            if self._pll_doubler_state is True:
                doubler = 2.0
            else:
                doubler = 1.0
            factor = self._pll_factor
            sfreq = self._ifreq * doubler * factor
        else:
            sfreq = self._ifreq
        return sfreq

    def set_ofreq(self, value):
        """Set output frequency on DDS.
        Take the input and output frequency as argument and set the adequat
        register in the DDS and return the actual output frequency (see
        _actual_ofreq()).
        output frequency may be (a bit) different than  The
        :param value: Output frequency value (float).
        :returns: Actual output frequency, -1 if process failed (float)
        """
        value = float(value)
        if self.is_connected() is False:
            self._ofreq = value
            return -1
        # Compute the Frequency Tuning Word (FTW).
        ftw = int((value * (1 << FTW_SIZE)) / self.get_sysfreq())
        # Prepare list of value to send to the 6 FTW0 registers.
        ftw_val_list = split_len(format(ftw, '012x'), 2)
        # Send values to the 6 FTW0 registers of the DDS.
        for idx, ftw_val in enumerate(ftw_val_list):
            if self.set_reg(self._regName2addr['Ftw0_5']-idx, int(ftw_val, 16)) != 0:
                return -1
        # Return the actual output frequency.
        self._ofreq = self._actual_ofreq(self.get_sysfreq(), ftw, FTW_SIZE)
        logging.debug("Set output frequency: %r", self._ofreq)
        return self._ofreq

    def get_ofreq(self):
        """Get output frequency of DDS.
        :returns: Output frequency of DDS (float).
        """
        if self.is_connected() is False:
            return self._ofreq
        # Get back the Frequency Tuning Word (FTW) from device.
        ftw = 0
        for idx in range(0, 6):
            ftw += self.get_reg(self._regName2addr['Ftw0_0']+idx) << (idx*8)
        self._ofreq = self._actual_ofreq(self.get_sysfreq(), ftw, FTW_SIZE)
        return self._ofreq

    def set_phy(self, value):
        """Set phase of output signal on DDS.
        Take the queried output phase (in degree) as argument and set
        the adequat register in the DDS.
        :param value: Output phase value (float).
        :returns: Actual output phase, -1 if process failed (float)
        """
        value = float(value)
        if self.is_connected() is False:
            self._phy = value
            return -1
        # Compute dphy ("delta phase")
        dphy = int((value * (1 << PHASE_SIZE)) / 360)
        # Prepare list of value to send to the dphy registers.
        dphy_val_list = split_len(format(dphy, '04x'), 2)
        # Send values to the dphy registers of the DDS.
        # TODO Search a better way to find reg address.
        for idx, val in enumerate(dphy_val_list):
            if self.set_reg(self._regName2addr['ADphase']-idx, int(val, 16)) != 0:
                return -1
        # Return the actual output phase
        adphy = self.actual_phy(dphy)
        logging.debug("Set phase: %r", adphy)
        return adphy

    def get_phy(self):
        """Get output phase of DDS.
        :returns:  Output phase of DDS (float).
        """
        # Get back the dphy register content
        dphy = 0
        for idx in range(0, 2):
            dphy += self.get_reg(self._regName2addr['ACphase']+idx) << (idx*8)
        # return the ofreq.
        return self.actual_phy(dphy)

    def set_amp(self, value):
        """Set amplitude tuning word of output signal on DDS.
        Take the input and output frequency as argument and set the adequat
        register in the DDS.
        :param value: Output amplitude value (int)
        :returns: fsc register value, -1 if process failed (int)
        """
        value = int(float(value))
        # If value is out of range, bound value and raise Warning.
        if not 0 <= value <= AMAX:
            value = bound_value(value, 0, AMAX)
            logging.warning("Amplitude value out of range")
            logging.info("Amplitude set to %r", value)
        if self.is_connected() is False:
            self._amp = value
            return -1
        # Prepare list of value to send to the fsc registers.
        fsc_val_list = split_len(format(value, '04x'), 2)
        # Send values to the fsc registers of the DDS.
        # TODO Search a better way to find reg address.
        for idx, val in enumerate(fsc_val_list):
            if self.set_reg(self._regName2addr['DacFsMsb']-idx, int(val, 16)) != 0:
                return -1
        # Return the amplitude value if transfert ok.
        logging.debug("Set output amplitude: %r", value)
        return value

    def get_amp(self):
        """Get output amplitude tuning word of DDS.
        :returns:  Output amplitude tuning of DDS (float).
        """
        # Get back the FSC register (Full Scale) register content
        fsc = 0
        for idx in range(0, 2):
            fsc += self.get_reg(self._regName2addr['DacFsLsb']+idx) << (idx*8)
        # Return the amplitude tuning word.
        return fsc

    def actual_phy(self, dphy):
        """Public function for specific AD9912 product.
        :param dphy: Phase register value (float).
        :returns: Actual output phase offset in degree (float).
        """
        return self._actual_phy(dphy, PHASE_SIZE)

    def set_hstl_output_state(self, state=False):
        """Set HSTL output state.
        :param state: - False  Disable HSTL output. (bool)
                      - True   Enable HSTL output.
        :returns: None
        """
        if self.is_connected() is not True:
            return
        reg = self.get_reg(self._regName2addr['PDEn'])
        if state is True:
            mask = ~(1 << 7)
            msg = reg & mask
        else:
            mask = 1 << 7
            msg = reg | mask
        self.set_reg(self._regName2addr['PDEn'], msg)
        logging.debug("Set HSTL output state to: %s", str(state))

    def get_hstl_output_state(self):
        """Get HSTL output state.
        :returns: HSTL output state (bool)
        """
        # Get back the Power Down and Enable register
        pd_en = self.get_reg(self._regName2addr['PDEn'])
        # return the HSTL output state
        hstl = pd_en & (1 << 7)
        return bool(hstl == 0)

    def set_cmos_output_state(self, state=False):
        """Set CMOS output state.
        :param state: - False  Disable CMOS output. (bool)
                      - True   Enable CMOS output.
        :returns: None
        """
        if self.is_connected() is not True:
            return
        reg = self.get_reg(self._regName2addr['PDEn'])
        if state is True:
            mask = 1 << 6
            msg = reg | mask
        else:
            mask = ~(1 << 6)
            msg = reg & mask
        self.set_reg(self._regName2addr['PDEn'], msg)
        logging.debug("Set CMOS output state to: %s", str(state))

    def get_cmos_output_state(self):
        """Get CMOS output state.
        :returns: CMOS output state (bool)
        """
        # Get back the Power Down and Enable register
        pd_en = self.get_reg(self._regName2addr['PDEn'])
        # return the CMOS output state
        cmos = pd_en & (1 << 6)
        if cmos == 0:
            retval = False
        else:
            retval = True
        return retval

    def set_pll_state(self, state=False):
        """Set PLL multiplier state.
        Note: A modification of the PLL state modify the output frequency.
        :param state: - False  Disable PLL. (bool)
                      - True   Enable PLL.
        :returns: None
        """
        if self.is_connected() is True:
            reg = self.get_reg(self._regName2addr['PDEn'])
            if state is True:
                mask = ~(1 << 4)
                msg = reg & mask
            else:
                mask = 1 << 4
                msg = reg | mask
            self.set_reg(self._regName2addr['PDEn'], msg)
        self._pll_state = state
        logging.debug("Set PLL state to: %s", str(state))

    def get_pll_state(self):
        """Get PLL state.
        :returns: PLL state (bool)
        """
        # Get back the Power Down and Enable register
        pd_en = self.get_reg(self._regName2addr['PDEn'])
        # return the PLL state
        doubler = pd_en & (1 << 4)
        return bool(doubler == 0)

    def set_pll_doubler_state(self, flag=False):
        """Set SysClk PLL doubler state.
        Note: A modification of the PLL doubler state modify the output
        frequency.
        :param flag: - False  Disable frequency doubler. (bool)
                     - True   Enable frequency doubler.
        :returns: None
        """
        if self.is_connected() is True:
            reg = self.get_reg(self._regName2addr['PllParam'])
            if flag is True:
                mask = 1 << 3
                msg = reg | mask
            else:
                mask = ~(1 << 3)
                msg = reg & mask
            self.set_reg(self._regName2addr['PllParam'], msg)
        self._pll_doubler_state = flag
        logging.debug("Set PLL doubler state to: %r", flag)

    def get_pll_doubler_state(self):
        """Get PLL doubler state.
        :returns: PLL doubler state (bool)
        """
        # Get back the Pll Parameters register
        pll_param = self.get_reg(self._regName2addr['PllParam'])
        # return the PLL doubler state
        doubler = pll_param & (1 << 3)
        return bool(doubler == 1 << 3)

    def set_pll_multiplier_factor(self, factor):
        """Set SysClk PLL multiplier factor.
        Note1: here we assign the overall multiplier factor, so the prescaler
        divider by 2 in the SysClk PLL multiplier block is include in the given
        factor.
        Note2: A modification of the PLL pultiplier factor modify the output
        frequency.
        :param factor: factor of PLL multiplier (between 4 to 66) (int)
        :returns: None
        """
        factor = bound_value(factor, 4, 66, "multiplier")
        if self.is_connected() is True:
            msg = factor / 2  # To take account of prescaler
            self.set_reg(self._regName2addr['Ndivider'], msg)
        self._pll_factor = factor
        logging.debug("Set PLL multiplier factor to: %r", factor)

    def get_pll_multiplier_factor(self):
        """Get SysClk PLL multiplier factor.
        Note that here we get the overall multiplier factor, so the prescaler
        divider by 2 in the SysClk PLL multiplier block is include in the
        returned factor.
        :returns: factor of PLL multiplier (between 4 to 66) (int)
        """
        n_divider = self.get_reg(self._regName2addr['Ndivider'])
        return n_divider

    def set_cp_current(self, value=0):
        """Set charge pump current value.
        :param value: charge pump current: - 0: 250 uA
                                           - 1: 375 uA
                                           - 2: off
                                           - 3: 125 uA
        :returns: None
        """
        if self.is_connected() is not True:
            return
        reg = self.get_reg(self._regName2addr['PllParam'])
        reg = reg & ~((1 << 1) + (1 << 0))
        msg = reg + value
        self.set_reg(self._regName2addr['PllParam'], msg)
        logging.debug("Set charge pump current to: %r", value)

    def get_cp_current(self):
        """Get charge pump current configuration value.
        Charge pump current: - 0: 250 uA
                             - 1: 375 uA
                             - 2: off
                             - 3: 125 uA
        :returns: charge pump current (int)
        """
        # Get back the Pll Parameters register
        pll_param = self.get_reg(self._regName2addr['PllParam'])
        # return the CP current config value.
        retval = pll_param & 3
        return retval

    def set_vco_range(self, value=None):
        """Set VCO range.
        :param value: vco range: - 0 = low range (int)
                                 - 1 = high range
                                 - others = autorange
        :returns: None
        """
        if self.is_connected() is not True:
            return
        reg = self.get_reg(self._regName2addr['PllParam'])
        if value == 0:
            # Disable autorange
            mask = ~(1 << 7)
            reg = reg & mask
            # Set low range
            mask = ~(1 << 2)
            msg = reg & mask
        elif value == 1:
            # Disable autorange
            mask = ~(1 << 7)
            reg = reg & mask
            # Set high range
            mask = 1 << 2
            msg = reg | mask
        else:
            mask = 1 << 7
            msg = reg | mask
        self.set_reg(self._regName2addr['PllParam'], msg)
        logging.debug("Set VCO range to: %r", value)

    def get_vco_range(self):
        """Get VCO range configuration value.
        vco range: - 0 = low range
                   - 1 = high range
                   - 2 = autorange
        :returns: vco range (int)
        """
        # Get back the Pll Parameters register
        pll_param = self.get_reg(self._regName2addr['PllParam'])
        autorange = pll_param & (1 << 7)
        vcorange = pll_param & (1 << 2)
        # return the VCO range config value.
        if autorange == 1 << 7:
            retval = 2
        else:
            retval = vcorange >> 2
        return retval

    def set_hstl_doubler_state(self, flag=False):
        """Set HSTL output doubler state: multiply HSTL frequecy by 2.
        :param flag: - False  Disable frequency doubler. (bool)
                     - True   Enable frequency doubler.
        :returns: None
        """
        if self.is_connected() is not True:
            return
        # Set HSTL driver register
        reg = self.get_reg(self._regName2addr['HstlDrv'])
        if flag is True:
            mask = 1 << 0
            msg = reg | mask
            mask = ~(1 << 1)
            msg = reg & mask
        else:
            mask = 1 << 1
            msg = reg | mask
            mask = ~(1 << 0)
            msg = reg & mask
        self.set_reg(self._regName2addr['HstlDrv'], msg)
        # Set Power-Down and Enable register
        reg = self.get_reg(self._regName2addr['PDEn'])
        if flag is True:
            mask = 1 << 5
            msg = reg | mask
        else:
            mask = ~(1 << 5)
            msg = reg & mask
        self.set_reg(self._regName2addr['PDEn'], msg)
        #
        logging.debug("Set HSTL doubler state to: %r", flag)

    def get_hstl_doubler_state(self):
        """Get HSTL doubler state.
        :returns: HSTL doubler state (bool)
        """
        # Get back the Pll Parameters register
        pll_param = self.get_reg(self._regName2addr['HstlDrv'])
        # return the PLL doubler state
        doubler = pll_param & (1 << 1) & (1 << 0)
        return bool(doubler == 2)

    def set_led(self, flag=False):
        """This gives the user control over the USB Status LED on the board.
        :param flag: - False  Stop the firmware controled LED Flashing and
                              gives control to the software.
                     - True   Resume the firmware controled LED Flashing.
        :returns: None
        """
        if self.is_connected() is not True:
            return
        if flag is True:
            msg = array('B', (0x0F, 0x1))
        else:
            msg = array('B', (0x0F, 0x0))
        try:
            self._fx2.write_ctrl(msg)
        except usb.core.USBError as ex:
            logging.error("Could set LED blink: %r", ex)

    def set_reg(self, address, value):
        """Set register value @ address.
        :param address: Address of the register to write (int).
        :param value: Value to write (int).
        :returns: 0 if ok else -1 (int)
        """
        value = int(value)
        # Build a list of binary value corresponding to the instruction
        bword = self._build_instruct(0, 0, address, value)
        bword = (bin(bword)).replace('0b', '')
        bword = list(bword)
        # Build the formated instruction message
        bword.reverse()
        msg = array('B', [0 for i in range(24)])
        for idx, val in enumerate(bword):
            if val == '1':
                msg[len(msg)-1-idx] = 1
        write_flag = False
        while write_flag is False:
            # Write real data
            try:
                self._fx2.write(msg)
            except usb.core.USBError as ex:
                logging.error("Write msg %r failed: %r", msg, ex)
                return -1
            # End write cycle (Set value to portA's pins of FX2 device)
            try:
                end = array('B', (0x01, 0x0B))
                self._fx2._ep_out.write(end)
                end = array('B', (0x01, 0x03))
                self._fx2._ep_out.write(end)
            except usb.core.USBError as ex:
                logging.error("Set %r @address %r failed: %r",
                              value, address, ex)
                return -1
            reg_value = self.get_reg(address)
            if reg_value == value:
                write_flag = True
        return 0


    def get_reg(self, address):
        """Get register value @ address.
        :param address: address of the register to write (int).
        :returns: register value (int).
        """
        # Sets the number of bytes that must be read back to 8 (0x8).
        init = array('B', (0x07, 0x00, 0x8))
        try:
            self._fx2._ep_out.write(init)
        except usb.core.USBError as ex:
            logging.error("Get register cycle failed: %r", ex)
            return -1
        # Query register reading operation
        bword = self._build_instruct(1, 0, address)
        bword = (bin(bword)).replace('0b', '')
        bword = list(bword)
        bword.reverse()
        msg = array('B', [0 for i in range(16)])
        for idx, val in enumerate(bword):
            if val == '1':
                msg[len(msg)-1-idx] = 1
        try:
            self._fx2.write(msg)
        except usb.core.USBError as ex:
            logging.error("Write msg %r failed: %r", msg, ex)
            return -1
        # Read result.
        try:
            raw_data = self._fx2.read()
        except usb.core.USBError as ex:
            logging.error("Read raw_data failed: %r", ex)
            return -1
        # Disable single byte read mode (readback mode).
        end = array('B', (0x04, 0x00))
        try:
            self._fx2._ep_out.write(end)
        except usb.core.USBError as ex:
            logging.error("Get register cycle failed: %r", ex)
            return -1
        # Build the register value from the raw data.
        result = 0
        for idx, val in enumerate(raw_data):
            result += int(val) << len(raw_data)-1-idx
        return result

    @staticmethod
    def _build_instruct(rw, bytes2send, reg_addr, reg_value=0):
        """Builds an instruction.
        Take arguments to form adequat instruction.
        Return an array containing the serie of instruction bits
        (i.e. each case of array contains '0' or '1').
        Details:   Prototype of transmission on AD9912 serial bus:
             16 bits header    | (up to) 4 bytes data
        R/W | W1:W0 |  A12:A0  |     ............
        ex: read FTW0_0 register
         1  |  0:0  |  0x0000  |
         => 0x8000
        ex: read Part ID (2 bytes)
         1  |  0:1  |  0x0002  |
         => 0xA002
        :param rw: 1 = read, 0 = Write (int).
        :param bytes2Send: 0=1Byte, 1=2Bytes, 2=3Bytes, 3=4Bytes or more (int).
        :param regAddr: enter the register address to write to (int).
        :param regValue: only if register writing is requested (int).
        :returns: instruction (int)
        """
        if rw == 1:
            rw = 1 << 15
            bytes2send = bytes2send << 13
            instruct = rw + bytes2send + reg_addr
        else:
            bytes2send = bytes2send << 21
            reg_addr = reg_addr << 8
            instruct = rw + bytes2send + reg_addr + reg_value
        return instruct

    @staticmethod
    def _actual_ofreq(ifreq, ftw, rsize):
        """Return the actual output frequency.
        Due to the resolution of the DDS, the actual output frequency
        may be (a bit) different than the queried.
        :param ftw: Frequency Tuning Word register value (int).
        :param rsize: Size of register (int).
        :returns: Actual output frequency (float).
        """
        return (ftw * ifreq) / (1 << rsize)

    @staticmethod
    def _actual_phy(dphy, bit):
        """Return the actual output phase.
        Due to the resolution of the DDS, the actual output phase
        may be (a bit) different than the queried.
        :param dphy: Phase register value (int).
        :param bit: Number of bits used for the phase resolution (int)
        :returns: Actual output phase offset in degree (float).
        """
        return float(360 * dphy) / (1 << bit)


# =============================================================================
if __name__ == '__main__':
    # Setup logger
    LOG_FORMAT = '%(asctime)s %(levelname)s %(filename)s (%(lineno)d): ' \
        + '%(message)s'
    logging.basicConfig(format=LOG_FORMAT, level=logging.DEBUG)

    IFREQ = 10000000.0
    OFREQ = 1000000.0
    DDS = Ad9912Dev()
    DDS.set_ofreq(OFREQ)
    DDS.set_ifreq(IFREQ)



# -*- coding: utf-8 -*-

"""package ad9xdds
author    Benoit Dubois
copyright FEMTO ENGINEERING, 2020
license   GPL v3.0+
brief     API to control AD9912 DDS development board with PyQt facilities.
details   Class derived Ad9912Dev class to add PyQt facilities
          i.e. signals/slots facilities
"""

from PyQt5.QtCore import QObject, pyqtSignal
from ad9xdds.ad9912dev import Ad9912Dev


# =============================================================================
class QAd9912Dev(QObject, Ad9912Dev):
    """Class derived from Ad9912Dev class to add PyQt facilities.
    """

    ifreqUpdated = pyqtSignal(float)
    ofreqUpdated = pyqtSignal(float)
    phaseUpdated = pyqtSignal(float)
    ampUpdated = pyqtSignal(int)
    pllStateUpdated = pyqtSignal(bool)
    pllDoublerUpdated = pyqtSignal(bool)
    pllFactorUpdated = pyqtSignal(int)

    def __init__(self, **kwargs):
        QObject.__init__(self, **kwargs)
        Ad9912Dev.__init__(self, **kwargs)

    def connect(self, **kwargs):
        return Ad9912Dev.connect(self, **kwargs)

    def set_ifreq(self, ifreq):
        Ad9912Dev.set_ifreq(self, ifreq)
        self.ifreqUpdated.emit(ifreq)

    def set_ofreq(self, ofreq):
        aofreq = Ad9912Dev.set_ofreq(self, ofreq)
        self.ofreqUpdated.emit(aofreq)
        return aofreq

    def set_phy(self, phy):
        aphy = Ad9912Dev.set_phy(self, phy)
        self.phaseUpdated.emit(aphy)
        return aphy

    def set_amp(self, amp):
        aamp = Ad9912Dev.set_amp(self, amp)
        self.ampUpdated.emit(aamp)
        return aamp

    def set_pll_state(self, state):
        Ad9912Dev.set_pll_state(self, state)
        self.pllStateUpdated.emit(state)

    def set_pll_doubler_state(self, flag):
        Ad9912Dev.set_pll_doubler_state(self, flag)
        self.pllDoublerUpdated.emit(flag)

    def set_pll_multiplier_factor(self, factor):
        Ad9912Dev.set_pll_multiplier_factor(self, factor)
        self.pllFactorUpdated.emit(factor)



# -*- coding: utf-8 -*-

"""package ad9xdds
author    Benoit Dubois
copyright FEMTO ENGINEERING, 2020-2025
license   GPL v3.0+
brief     API to control AD9915 DDS development board through USB to SPI
          adapter.
details   AD9915 development board is configurable by USB through
          FTDI 232H device, an USB to SPI transceiver. Class allows
          configuration of principle DDS parameters (frequency, phase,
          amplitude, PLL handling, output handling...).


Warning: correct installation of pyftdi also need to create a udev
configuration file:
- get it at https://eblot.github.io/pyftdi/installation.html
or
- use script install.sh available in archive firm_manager.tar.gz in the 'data'
directory of this package.


To connect to FTDI device use URL scheme defined as:
    ftdi://[vendor][:[product][:serial|:bus:address|:index]]/interface
With FTDI 232H device:
    ftdi://ftdi:232h:[:serial|:bus:address|:index]]/1
!! Warning !! bus and address must be in hexadecimal


FTDI UMR232H cable connection:
    UMR232H                              AD9915 Dev Board (P101)
    Black (GND)    <------------------>  GND
    Brown (CS0)    <------------------>  MPI00 (CSB)
    Orange (SCLK)  <------------------>  MPI01 (SCLK)
    Yellow (MOSI)  <------------------>  MPI02 (SDIO)
    Green (MISO)   <------------------>  MPI03 (SDO)
    Grey (GPIO4)   <------------------>  IO_UPDATE (GPLIO0)
    Purple (GPIO5) <------------------>  RESET_BUFF (GPLIO1)

    Profile selection (currently not implemented (P102))
    PS0-BUF <---> GND
    PS1-BUF <---> GND
    PS2-BUF <---> GND

Configuration of AD9915 Dev Board jumpers:
    Disable USB communication
    P203 <---> Vcc (Disable)
    P204 <---> Vcc (Disable)
    P205 <---> Vcc (Disable)

    Set serial programming mode (P101) (Datasheet AD9915 p.28):
    IOCFG3       <---> GND
    IOCFG2       <---> GND
    IOCFG1       <---> GND
    IOCFG0       <---> Vdd

    Others jumpers configuration (P101)
    SYNC_IO_BUFF <---> GND  (Disable I/O reset)

    Others jumpers configuration (P102)
    EXTPDCTL-BUF <---> GND
    RESET-BUF    <---> OPEN
    DROVR-BUF    <---> GND
    DRHOLD-BUF   <---> GND
    DRCTL-BUF    <---> GND

WARNING: To avoid problems when disconnecting the ftdi cable from USB port,
you need to add 100 kOhms pull down resistor to IOUPDATE-BUFF and RESET-BUF
lines. For the IOUPDATE-BUFF line you can add the resitor on R136B pads or
directly behind the P101 connector. For the RESET-BUF line add the resistor
directly behind the P101 connector or use a "resistor jumper" on the P102/P100
connector.
"""

import logging
import time
import fractions
from decimal import Decimal
import pyftdi.ftdi as ftdi
import pyftdi.spi as spi

DEBUG = False

SPI_CLOCK_FREQUENCY = 100000
SPI_MODE = 0

IFMAX = 2500000000    # Input maximum frequency (Hz)
FTW_SIZE = 32         # Frequency Tuning Word register size (bit)
PHASE_SIZE = 16       # Phase register size (bit)
DAC_OUT_SIZE = 12     # Output DAC resolution (bit)
AMAX = (1 << DAC_OUT_SIZE) - 1  # Output maximum amplitude (a.u.)

REGNAME2ADDR = {'CFR1': 0x0, 'CFR2': 0x1, 'CFR3': 0x2, 'CFR4': 0x3,
                'DigRampLowerLimit': 0x4, 'DigRampUpperLimit': 0x5,
                'RisingDigRampStepSize': 0x6, 'FallingDigRampStepSize': 0x7,
                'DigRampRate': 0x8,
                'LowerFreqJump': 0x9, 'UpperFreqJump': 0xA,
                'P0Ftw': 0xB, 'P0PhaseAmp': 0xC,
                'P1Ftw': 0xD, 'P1PhaseAmp': 0xE,
                'P2Ftw': 0xF, 'P2PhaseAmp': 0x10,
                'P3Ftw': 0x11, 'P3PhaseAmp': 0x12,
                'P4Ftw': 0x13, 'P4PhaseAmp': 0x14,
                'P5Ftw': 0x15, 'P5PhaseAmp': 0x16,
                'P6Ftw': 0x17, 'P6PhaseAmp': 0x18,
                'P7Ftw': 0x19, 'P7PhaseAmp': 0x1A,
                'USR0': 0x1B}

OPMODE = {'profile': 0,  # single tone or modulation
          'drg' : 1 ,
          'parallel': 2,
          'modulus': 3}

CFR1 = {'OSK_EN': 8}

CFR2 = {'PROFILE_EN': 23,
        'PARALLEL_EN': 22,
        'DRG_EN': 19,
        'MODULUS_EN': 16}

MAX_REQUEST = 10


# =============================================================================
def get_interfaces(pid=0x0403, vid=0x6014):
    """Get available interface of UMR232 FTDI devices.
    Return a list of UsbDeviceDescriptor, a named tuple with the following
    fields: vid, pid, bus, address, sn, index, description'.
    :returns: list of UsbDeviceDescriptor (list of NamedTuple)
    """
    interfaces = ftdi.Ftdi().find_all([(pid, vid),])
    return [interface[0] for interface in interfaces]


# =============================================================================
class ConnectionError(Exception):
    pass


# =============================================================================
class InvalidState(Exception):
    """Defined exception when DDS seems in an unknown state"""
    pass


# =============================================================================
class Ad9915Dev():
    """Class representing AD9915 development board through USB to SPI adapter.
    Currently class use profile mode P0 only
    """

    def __init__(self, ifreq=IFMAX):
        """The constructor.
        :param ifreq: Current input frequency in Hz (float)
        :returns: None
        """
        self._url = None
        # Init properties related to FTDI device
        self._ctrl = None  # SPI controller
        self._spi = None  # SPI bus
        self._gpio = None  # Others signals needed to handle DDS
        # Init class properties
        self._ifreq = ifreq
        self._profile = 0
        self._ofreq = None
        self._phy = None
        self._amp = None
        self._pll_state = None
        self._pll_doubler_state = None
        self._pll_factor = None
        self._opmode = None

    def __del__(self):
        if self._ctrl is None:
            return
        if self.is_connected() is True:
            self.disconnect()
        if self._ctrl is not None:
            self._ctrl.terminate()

    def connect(self, url):
        """Connection process.
        :param url: FTDI url like 'ftdi://ftdi:232h:FT0GPCDF/0' (str)
        :returns: None
        """
        if self._url is None and url is None:
            logging.error("No device defined before connection")
            return
        self._url = url
        self._ctrl = spi.SpiController()
        self._ctrl.configure(self._url, debug=DEBUG)
        self._spi = self._ctrl.get_port(cs=0)
        self._spi.set_frequency(SPI_CLOCK_FREQUENCY)
        self._spi.set_mode(SPI_MODE)
        print("SPI configured")
        self._gpio = self._ctrl.get_gpio()
        ## GPIOL0 (b4) and GPIOL1 (b5) are outputs (='1')
        ## GPIOL0 is IO_UPDATE and GPIOL1 is MASTER_RESET
        self._gpio.set_direction(0x30, 0x30)
        ## Init IO_update and reset lines to 0
        self._gpio.write(0x00)
        # Configure/Init DDS device
        ## Reset: needed at power-up (see datasheet p.40)
        self._master_reset()
        ## Default + SDIO input only + POWERDOWN DIGITAL, DAC and REFCLK part
        self.set_reg_w16(REGNAME2ADDR['CFR1'], [0x00, 0x01, 0x00, 0xEA])
        ## Default + SDIO input only + POWERUP DIGITAL, DAC and REFCLK part
        self.set_reg_w16(REGNAME2ADDR['CFR1'], [0x00, 0x01, 0x00, 0x0A])
        ## Default + Enable profile mode
        self.set_reg_w16(REGNAME2ADDR['CFR2'], [0x00, 0x80, 0x09, 0x00])
        ## DAC cal: needed at power-up (see datasheet p.40)
        self.dac_calibration()
        self._get_dds_parameters()
        return True

    def disconnect(self):
        """Disconnect from FTDI device.
        :returns: None
        """
        if self._ctrl is not None:
            if not self._ctrl.ftdi.is_connected:
                return
            self._ctrl.ftdi.close()

    def is_connected(self):
        """Return True if interface to DDS board is ready else return False.
        """
        if self._ctrl is not None:
            if self._ctrl.ftdi.is_connected:
                return True
        return False

    def _get_dds_parameters(self):
        """Get parameters from DDS.
        """
        # !warning! parameter order initialisation is important
        '''
        if self._pll_state is None:
            self._pll_state = self.get_pll_state()
        else:
            self.set_pll_state(self._pll_state)
        if self._pll_doubler_state is None:
            self._pll_doubler_state = self.get_pll_doubler_state()
        else:
            self.set_pll_doubler_state(self._pll_doubler_state)
        if self._pll_factor is None:
            self._pll_factor = self.get_pll_multiplier_factor()
        else:
            self.set_pll_multiplier_factor(self._pll_factor)
        '''
        if self._ofreq is None:
            self._ofreq = self.get_ofreq()
        else:
            self.set_ofreq(self._ofreq)
        if self._phy is None:
            self._phy = self.get_phy()
        else:
            self.set_phy(self._phy)
        if self._amp is None:
            self._amp = self.get_amp()
        else:
            self.set_amp(self._amp)
        if self._opmode is None:
            try:
                self._opmode = self.get_operation_mode()
            except InvalidState as ex:
                self.set_operation_mode(OPMODE['modulus'])
                logging.error("No default operation mode defined, set modulus mode")
        else:
            self.set_operation_mode(self._opmode)

    def get_reg_w16(self, address, length=4):
        """Get value of a register in 8 bits word size.
        :param address: register address (int)
        :param length: size of register in byte (int)
        :returns: register value in 16 bits word size (list of int)
        """
        msg = [0x80 + address]
        self._spi.write(msg, stop=False)
        self._io_update()
        retval = self._spi.read(length, start=False)
        return list(retval)

    def set_reg_w16(self, address, values, length=4):
        """Set value of a register in 8 bits word size.
        :param address: register address (int)
        :param value: register value list to set (list of int)
        :param length: size of register in byte (int)
        :returns: None
        """
        msg = [address]
        for value in values:
            msg.append(value)
        request = 0
        while request <= MAX_REQUEST:
            self._spi.write(msg, stop=False)
            self._io_update()
            self._spi.write(out=[], stop=True)
            reg_value = list(self.get_reg_w16(address, length))
            if reg_value == values:
                return
            else:
                logging.error("set_reg_w16 %r to %r error (get %r), attempt %r",
                              address, values, reg_value, request)
                request += 1
        raise ConnectionError("Could not set register {}".format(address))

    def get_reg(self, address, length=4):
        """Get register value.
        :param address: register address (int)
        :param length: size of register in byte (int)
        :returns: register value (int)
        """
        reg_value = self.get_reg_w16(address, length)
        retval = 0
        for i, val in enumerate(reg_value):
            retval += val << (8*(len(reg_value)-i-1))
        return retval

    def set_reg(self, address, value, length=4):
        """Set register value.
        :param address: register address (int)
        :param value: register value to set (int)
        :param length: size of register in byte (int)
        :returns: None
        """
        values = []
        for i in range(length-1, -1, -1):
            values.append((value >> (i * 8)) & 0xff)
        self.set_reg_w16(address, values)

    def get_reg_bit(self, address, idx):
        """Get value of a specific bit of a register.
        :param address: register address (int)
        :param idx: index position of bit to read (int)
        :returns: bit value, i.e. 0 or 1 (int)
        """
        reg_value = self.get_reg(address)
        return self.get_bit_value(reg_value, idx)

    def set_reg_bit(self, address, idx, value):
        """Set value of a specific bit of a register.
        :param address: register address (int)
        :param idx: index position of bit to read (int)
        :param value: bit value, i.e. 0 or 1 (int)
        :returns: None
        """
        reg_value = self.get_reg(address)
        # If bit value already set, stop update
        if self.get_bit_value(reg_value, idx) == value:
            return
        mask = 1 << idx
        self.set_reg(address, self.set_bit_value(reg_value, idx, value))

    @staticmethod
    def get_bit_value(n, idx):
        """Get value of a specific bit 'idx' in integer 'n'.
        :param n: integer under test  (int)
        :param idx: index position of bit to read (int)
        :returns: bit value, i.e. 0 or 1 (int)
        """
        return (n >> idx) & 1

    @staticmethod
    def set_bit_value(n, idx, value):
        """Set value 'value' of a specific bit 'idx' in integer 'n'.
        :param n: integer under test  (int)
        :param idx: index position of bit to read (int)
        :param value: bit value, i.e. 0 or 1 (int)
        :returns: integer 'n' updated (int)
        """
        mask = 1 << idx
        if value == 0:
            return n & ~mask
        else:
            return n | mask

    def dac_calibration(self):
        """DAC calibration, needed after each power-up and every time REF CLK or
        the internal system clock is changed.
        :returns: None
        """
        # Calibration 1/2: enable
        self.set_reg_w16(REGNAME2ADDR['CFR4'], [0x01, 0x05, 0x21, 0x20])
        # Wait for calibration (max calibration time is 188 us)
        time.sleep(0.001)
        # Calibration 2/2: disable
        self.set_reg_w16(REGNAME2ADDR['CFR4'], [0x00, 0x05, 0x21, 0x20])

    def set_profile(self, profile):
        """Set profile currently in use. Curently not implemented.
        :param profile: Select profile in use between 0 to 7 (int)
        :returns: None
        """
        if profile not in range(0, 8):
            raise ValueError("Profile must be in range 0 to 7, here: ",
                             profile)
        ## self._profile = profile

    def get_profile(self):
        """Get profile currently in use. Curently not implemented.
        :returns: profile currently in use (int)
        """
        return self._profile

    def set_operation_mode(self, mode: int) -> None:
        """Configure operation mode from:
        - Single tone (0)
        - Profile modulation (1)
        - Digital ramp modulation (linear sweep) (2) (not implemented)
        - Parallel data port modulation (3) (not implemented)
        - Programmable modulus mode (4)
        :param mode: index of operation mode (int)
        :returns: None
        """
        if mode == OPMODE['profile']:
            self.set_reg_bit(REGNAME2ADDR['CFR2'], CFR2['MODULUS_EN'], 0)
            self.set_reg_bit(REGNAME2ADDR['CFR2'], CFR2['PROFILE_EN'], 1)
            self.set_reg_bit(REGNAME2ADDR['CFR2'], CFR2['DRG_EN'], 0)
            self.set_reg_bit(REGNAME2ADDR['CFR1'], CFR1['OSK_EN'], 1)
        elif mode == OPMODE['drg']:
            self.set_reg_bit(REGNAME2ADDR['CFR2'], CFR2['MODULUS_EN'], 0)
            self.set_reg_bit(REGNAME2ADDR['CFR2'], CFR2['DRG_EN'], 1)
            self.set_reg_bit(REGNAME2ADDR['CFR1'], CFR1['OSK_EN'], 1)
        elif mode == OPMODE['parallel']:
            self.set_reg_bit(REGNAME2ADDR['CFR2'], CFR2['PARALLEL_EN'], 1)
            self.set_reg_bit(REGNAME2ADDR['CFR2'], CFR2['MODULUS_EN'], 0)
            self.set_reg_bit(REGNAME2ADDR['CFR2'], CFR2['PROFILE_EN'], 0)
            self.set_reg_bit(REGNAME2ADDR['CFR2'], CFR2['DRG_EN'], 0)
            self.set_reg_bit(REGNAME2ADDR['CFR1'], CFR1['OSK_EN'], 0)
        elif mode == OPMODE['modulus']:
            # Enable modulus mode bit
            self.set_reg_bit(REGNAME2ADDR['CFR2'], CFR2['MODULUS_EN'], 1)
            # Enable digital ramp modulation mode bit
            self.set_reg_bit(REGNAME2ADDR['CFR2'], CFR2['DRG_EN'], 1)
            # Enable OSK
            self.set_reg_bit(REGNAME2ADDR['CFR1'], CFR1['OSK_EN'], 1)
        else:
            pass
        self._opmode = mode

    def get_operation_mode(self) -> int:
        """Get operation mode status.
        :returns: operation mode (int)
        """
        paren = self.get_reg_bit(REGNAME2ADDR['CFR2'], CFR2['PARALLEL_EN'])
        moden = self.get_reg_bit(REGNAME2ADDR['CFR2'], CFR2['MODULUS_EN'])
        profen = self.get_reg_bit(REGNAME2ADDR['CFR2'], CFR2['PROFILE_EN'])
        drgen = self.get_reg_bit(REGNAME2ADDR['CFR2'], CFR2['DRG_EN'])
        osken = self.get_reg_bit(REGNAME2ADDR['CFR1'], CFR1['OSK_EN'])
        opmode_reg = f"{paren}{moden}{profen}{drgen}{osken}"
        match opmode_reg:
            case "00101" | "10101":
                opmode = OPMODE['profile']
            case "00011" | "00111" | "10011" | "10111":
                opmode = OPMODE['drg']
            case "10000":
                opmode = OPMODE['parallel']
            case "01011" | "01111" | "11011" | "11111":
                opmode = OPMODE['modulus']
            case _:
                raise InvalidState("Operation mode undefined")
        self._opmode = opmode
        return opmode

    def get_sysfreq(self):
        """Get system frequency.
        Currently, does not support PLL.
        :returns: Current system frequency value (float)
        """
        '''if self._pll_state is True:
            if self._pll_doubler_state is True:
                doubler = 2.0
            else:
                doubler = 1.0
            factor = self._pll_factor
            sfreq = self._ifreq * doubler * factor
        else:
            sfreq = self._ifreq
        return sfreq
        '''
        return self._ifreq

    def set_ifreq(self, value):
        """Set input frequency.
        :param value: Input frequency value (float)
        :returns: None
        """
        value = float(value)
        self._ifreq = value
        self.dac_calibration()  # Needed!
        logging.debug("Set input frequency: %r", value)
        # Update DDS output frequency because ofreq = f(ifreq)
        self.set_ofreq(self._ofreq)
        return value

    def get_ifreq(self):
        """Get input frequency.
        :returns: Current input frequency value (float)
        """
        return self._ifreq

    @staticmethod
    def compute_modulus_parameters(ifreq, ofreq):
        """Compute modulus parameters values given ifreq and ofreq values.
        Even if input parameters can be of float type, input parameters
        are preferably of string type to avoid casting rounding error.
        Return FTW, A and B parameters.
        :param ifreq: Output frequency value (str).
        :param ofreq: Output frequency value (str).
        :returns: FTW, A and B parameters (int, int, int)
        """
        ifreq = fractions.Fraction(ifreq)
        ofreq = fractions.Fraction(ofreq)
        (m, n) = fractions.Fraction(ofreq, ifreq).as_integer_ratio()
        ftw = int(m * 2**32 / n)
        y = 2**32 * m - ftw * n
        (a, b) = fractions.Fraction(y, n).as_integer_ratio()
        return ftw, a, b

    def set_ofreq_fine(self, value):
        """Set output frequency using modulus operation mode of DDS.
        Note: Do not forget to set operation mode to 'modulus'.
        Even if input parameters can be of float type, input parameters
        are preferably of string type to avoid casting rounding error.
        Returns output frequency.
        :param value: Output frequency value (str).
        :returns: Actual output frequency (float)
        # """
        if self.is_connected() is False:
            logging.error("Device is not connected.")
            return None
        ifreq = Decimal(self.get_sysfreq())
        ofreq = Decimal(value)
        ftw, a, b = self.compute_modulus_parameters(ifreq, ofreq)
        self.set_reg(0x04, ftw)
        self.set_reg(0x06, a)
        self.set_reg(0x05, b)
        self._ofreq = self._actual_ofreq_fine(ifreq, ftw, a, b)
        logging.debug("Set fine output frequency: %r", self._ofreq)
        return self._ofreq

    def get_ofreq_fine(self):
        ifreq = self.get_sysfreq()
        ftw = self.get_reg(0x04)
        a = self.get_reg(0x06)
        b = self.get_reg(0x05)
        self._ofreq = self._actual_ofreq_fine(ifreq, ftw, a, b)
        return self._ofreq

    def set_ofreq(self, value, profile=None):
        """Set output frequency to current DDS profile if profile parameter is
        None or set output frequency of requested DDS profile.
        Return the actual output frequency (see _actual_ofreq() method).
        :param value: Output frequency value (float).
        :param profile: Profile to update between 0 to 7 (int)
        :returns: Actual output frequency (float)
        """
        if self._opmode == OPMODE['modulus']:
            return self.set_ofreq_fine(value)

        if self.is_connected() is False:
            logging.error("Device is not connected.")
            return None
        if profile not in range(0, 8) and profile is not None:
            raise ValueError("Profile not in range 0 to 7, here:", profile)
        if profile is None:
            profile = 0
        ofreq = float(value)
        regname = 'P{:d}Ftw'.format(profile)
        ftw = int((ofreq * (1 << FTW_SIZE)) / self.get_sysfreq())
        self.set_reg(REGNAME2ADDR[regname], ftw)
        self._ofreq = self._actual_ofreq(self.get_sysfreq(), ftw, FTW_SIZE)
        logging.debug("Set output frequency: %r", self._ofreq)
        return self._ofreq

    def get_ofreq(self, profile=None):
        """Get output frequency of current DDS profile if profile parameter is
        None or return output frequency of requested DDS profile.
        :param profile: Profile output frequency requested (int)
        :returns: Output frequency of DDS profile (float).
        """
        if self._opmode == OPMODE['modulus']:
            return self.get_ofreq_fine()

        if self.is_connected() is False:
            logging.error("Device is not connected.")
            return None
        if profile not in range(0, 8) and profile is not None:
            raise ValueError("Profile not in range 0 to 7, here:", profile)
        if profile is None:
            profile = 0
        regname = 'P{:d}Ftw'.format(profile)
        ftw = self.get_reg(REGNAME2ADDR[regname])
        self._ofreq = self._actual_ofreq(self.get_sysfreq(), ftw, FTW_SIZE)
        return self._ofreq

    def set_phy(self, value, profile=None):
        """Set phase of output signal on DDS.
        Take the queried output phase (in degree) as argument and set
        the adequat register in the DDS.
        :param value: Output phase value (float).
        :param profile: Profile to update between 0 to 7 (int)
        :returns: Actual output phase (float)
        """
        if self.is_connected() is False:
            logging.error("Device is not connected.")
            return None
        if profile not in range(0, 8) and profile is not None:
            raise ValueError("Profile not in range 0 to 7, here:", profile)
        if profile is None:
            profile = 0
        value = float(value)
        regname = 'P{:d}PhaseAmp'.format(profile)
        # Phase and amplitude output is handled by a common register.
        # Before writing new phase, we need to get current amplitude
        # and take care to not reset its value.
        reg_value = self.get_reg_w16(REGNAME2ADDR[regname])
        asf_list = reg_value[2:]
        phy = int((value * (1 << PHASE_SIZE)) / 360)
        phy_list = self._int_2_byte_list(phy, 2)
        msg = asf_list + phy_list
        self.set_reg_w16(REGNAME2ADDR[regname], msg)
        phy = self._actual_phy(phy, PHASE_SIZE)  # Return the actual phase
        logging.debug("Set phase: %r", phy)
        return phy

    def get_phy(self, profile=None):
        """Get output phase of profile..
        :param profile: Profile phase requested (int)
        :returns: Output phase of DDS (float).
        """
        if self.is_connected() is False:
            logging.error("Device is not connected.")
            return None
        if profile not in range(0, 8) and profile is not None:
            raise ValueError("Profile not in range 0 to 7, here:", profile)
        if profile is None:
            profile = 0
        regname = 'P{:d}PhaseAmp'.format(profile)
        reg_value = self.get_reg_w16(REGNAME2ADDR[regname])
        # Extract phase value from register value.
        phy = (reg_value[2] << 8) + reg_value[3]
        return self._actual_phy(phy, PHASE_SIZE)  # return the actual phase.

    def set_amp(self, value, profile=None):
        """Set amplitude tuning word of output signal on DDS.
        Take the input and output frequency as argument and set the adequat
        register in the DDS.
        :param value: Output amplitude value (int)
        :param profile: Profile to update between 0 to 7 (int)
        :returns: fsc register value if transfert is ok (int)
        """
        if self.is_connected() is False:
            logging.error("Device is not connected.")
            return None
        if profile not in range(0, 8) and profile is not None:
            raise ValueError("Profile not in range 0 to 7, here:", profile)
        if profile is None:
            profile = 0
        value = int(float(value))
        # If value is out of range, bound value and raise Warning.
        if not 0 <= value <= AMAX:
            logging.warning("Amplitude value out of range: %r", value)
            value = self._bound_value(value, 0, AMAX)
        regname = 'P{:d}PhaseAmp'.format(profile)
        # Phase and amplitude output is handled by a common register.
        # Before writing new amplitude, we need to get current phase
        # and take care to not reset its value.
        reg_value = list(self.get_reg_w16(REGNAME2ADDR[regname]))
        phy_list = reg_value[:2]
        asf_list = self._int_2_byte_list(value, 2)
        msg = asf_list + phy_list
        self.set_reg_w16(REGNAME2ADDR[regname], msg)
        logging.debug("Set output amplitude: %r", value)
        return value

    def get_amp(self, profile=None):
        """Get output amplitude tuning word of DDS.
        :param profile: Profile phase requested (int)
        :returns:  Output amplitude tuning of DDS (float).
        """
        if self.is_connected() is False:
            logging.error("Device is not connected.")
            return None
        if profile not in range(0, 8) and profile is not None:
            raise ValueError("Profile not in range 0 to 7, here:", profile)
        if profile is None:
            profile = 0
        regname = 'P{:d}PhaseAmp'.format(profile)
        reg_value = self.get_reg_w16(REGNAME2ADDR[regname])
        # Extract amplitude value from register value.
        asf = (reg_value[0] << 8) + reg_value[1]
        return asf

    def set_output_state(self, state=False):
        """Set output state.
        :param state: - False  Disable output. (bool)
                      - True   Enable CMOS output.
        :returns: None
        """
        raise NotImplementedError

    def get_output_state(self):
        """Get output state.
        :returns: Output state (bool)
        """
        reg_value = self.get_reg(REGNAME2ADDR['CFR1'])
        output_state_bit = 16
        return self._check_bit_set(reg_value, output_state_bit)

    def set_pll_state(self, state=False):
        """Set PLL state.
        Note: A modification of the PLL state modify the output frequency.
        :param state: - False  Disable PLL. (bool)
                      - True   Enable PLL.
        :returns: None
        """
        reg_value = self.get_reg(REGNAME2ADDR['CFR3'])
        pll_state_bit = 2
        if state is True:
            reg_value |= 1 << pll_state_bit
        else:
            reg_value &= ~(1 << pll_state_bit)
        self.set_reg(REGNAME2ADDR['CFR3'], reg_value)

    def get_pll_state(self):
        """Get PLL state.
        :returns: PLL state (bool)
        """
        reg_value = self.get_reg(REGNAME2ADDR['CFR3'])
        pll_state_bit = 2
        return self._check_bit_set(reg_value, pll_state_bit)

    def is_pll_locked(self):
        """Return the internal PLL lock (to the REF CLK input signal) state.
        :returns: True if the internal PLL is locked else return False (bool)
        """
        reg_value = self.get_reg(REGNAME2ADDR['USR0'])
        pll_lock_bit = 24
        return self._check_bit_set(reg_value, pll_lock_bit)

    def set_pll_divider_factor(self, value):
        """Set PLL feedback divider value.
        :param factor: factor of PLL divider (between 20 to 510) (int)
        :returns: None
        """
        if self.is_connected() is False:
            logging.error("Device is not connected.")
            return None
        if not 20 <= value <= 510:
            logging.error("PLL divider factor value out of range: %r", value)
            value = self._bound_value(value, 20, 510)
        cfr3 = self.get_reg_w16(REGNAME2ADDR['CFR3'])
        cfr3[1] = value
        self.set_reg_w16(REGNAME2ADDR['CFR3'], cfr3)
        logging.debug("Set PLL divider factor: %r", value)
        return value

    def get_pll_divider_factor(self):
        """Get SysClk PLL divider factor.
        Note that here we get the overall divider factor, so the prescaler
        divider by 2 in the SysClk PLL divider block is include in the
        returned factor.
        :returns: factor of PLL divider (between 4 to 66) (int)
        """
        raise NotImplementedError

    def set_cp_current(self, value=0):
        """Set charge pump current value.
        :param value: charge pump current: - 0: 250 uA
                                           - 1: 375 uA
                                           - 2: off
                                           - 3: 125 uA
        :returns: None
        """
        raise NotImplementedError

    def get_cp_current(self):
        """Get charge pump current configuration value.
        Charge pump current: - 0: 250 uA
                             - 1: 375 uA
                             - 2: off
                             - 3: 125 uA
        :returns: charge pump current (int)
        """
        raise NotImplementedError

    def vco_calibration(self, value=None):
        """VCO calibration process.
        :param value: vco range: - 0 = low range (int)
                                 - 1 = high range
                                 - others = autorange
        :returns: None
        """
        raise NotImplementedError

    @staticmethod
    def _actual_ofreq(ifreq, ftw, rsize):
        """Return the actual output frequency.
        Due to the resolution of the DDS, the actual output frequency
        may be (a bit) different than the queried.
        :param ifreq: Intput frequency value (float).
        :param ftw: Frequency Tuning Word register value (int).
        :param rsize: Size of register (int).
        :returns: Actual output frequency (float).
        """
        return (ftw * ifreq) / (1 << rsize)

    @staticmethod
    def _actual_ofreq_fine(ifreq, ftw, a, b):
        """Return the actual output frequency.
        Due to the resolution of the DDS, the actual output frequency
        may be (a bit) different than the queried.
        :param ifreq: Intput frequency value (float).
        :param ftw: Frequency Tuning Word register value (int).
        :param a: Modulus parameter (int).
        :param b: Modulus parameter (int).
        :returns: Actual output frequency (float).
        """
        return Decimal(ifreq) * (Decimal(ftw) + Decimal(a)/Decimal(b))/2**32

    @staticmethod
    def _actual_phy(dphy, bit):
        """Return the actual output phase.
        Due to the resolution of the DDS, the actual output phase
        may be (a bit) different than the queried.
        :param dphy: Phase register value (int).
        :param bit: Number of bits used for the phase resolution (int)
        :returns: Actual output phase offset in degree (float).
        """
        return float(360 * dphy) / (1 << bit)

    @staticmethod
    def _int_2_byte_list(value, size):
        """Take an integer and split it in a list of the corresponding
        'size' 8 bits word (byte).
        For example:
            17179869 in hexadecimal = 0x10624dd
            0x10624dd splitted in byte = [0x01, 0x06, 0x24, 0xdd]
            [0x01, 0x06, 0x24, 0xdd] in base 10 = [1, 6, 36, 221]
            => _int_2_byte_list(17179869, 4) = [1, 6, 36, 221]
        :param value: an integer to split (int)
        ;param size: sier of output list (int)
        :returns: list of byte (list of int)
        """
        value_format = '0{:d}x'.format(size * 2)
        value = format(value, value_format)
        return [int(value[i:i+2], 16) for i in range(0, size*2, 2)]

    @staticmethod
    def _bound_value(value, vmin, vmax):
        """Check that a value is included in the range [min, max], if not the value
        is bounded to the range, ie:
        - if value < min  ->  min = value
        - if value > max  ->  max = value
        :param value: Value that is checked
        :param vmin: Minimum valid value.
        :param vmax: Maximum valid value.
        :returns: Bounded value.
        """
        if value < vmin:
            logging.warning("Parameter out of range (%f). Set to: %f",
                            value, vmin)
            return vmin
        if value > vmax:
            logging.warning("Parameter out of range (%f). Set to: %f",
                            value, vmax)
            return vmax
        return value

    @staticmethod
    def _check_bit_set(word, n):
        """Check if n-th bit of word is set or not.
        :param word: word to check (int)
        :param n: n-th bit of work to check (int)
        :returns: True is n-th bit is set else False (bool)
        """
        if word & (1 << n):
            return True
        return False

    def _io_update(self):
        """Generate IO update event: transfer written data from I/O buffer to
        the coresponding internal registers.
        :returns: None
        """
        self._gpio.write(0x10)
        time.sleep(0.001)
        self._gpio.write(0x0)

    def _master_reset(self):
        """Master reset: clears all memory elements and sets registers to
        default values. Required after power up.
        :returns: None
        """
        self._gpio.write(0x20)
        time.sleep(int(1 / (1000 * SPI_CLOCK_FREQUENCY)))
        self._gpio.write(0x0)


# =============================================================================
if __name__ == '__main__':
    """Basic hardware test.
    """
    LOG_FORMAT = '%(asctime)s %(levelname)s %(filename)s (%(lineno)d): ' \
        + '%(message)s'
    logging.basicConfig(format=LOG_FORMAT, level=logging.DEBUG)

    pylogger = logging.getLogger('pyftdi')
    pylogger.setLevel(logging.INFO)

    # List FTDI URL but exit imediately after:
    #print(spi.SpiController().configure('ftdi:///?'))

    devices = get_interfaces(0x0403, 0x6010)
    for dev in devices:
        print("{}, SN: {}, bus: {}, address:{}". \
              format(dev.description, dev.sn, dev.bus, hex(dev.address)))

    IFREQ = 0.5e9
    OFREQ = IFREQ * 0.1
    PHY = 0
    AMP = 4095

    DDS = Ad9915Dev()

    DDS.connect(url='ftdi://ftdi:232h/1')

    DDS.set_ifreq(IFREQ)

    DDS.set_phy(PHY)
    print("Get phase", DDS.get_phy())

    DDS.set_amp(AMP)
    print("Get amplitude", DDS.get_amp())

    DDS.set_ofreq(OFREQ)
    print("Get output frequency", DDS.get_ofreq())



# -*- coding: utf-8 -*-

""" ad9xdds/version.py """
__version__ = "0.5.10"

# 0.5.10 (2025/07/17): Fix rounding error with low resolution ofreq value:
#                      Use decimal.Decimal fixed point for arithmetic.
# 0.5.9  (02/04/2025): Correct set_ifreq() return value in AD9915Dev class.
# 0.5.8  (21/03/2025): AD9915: bugs correction.
# 0.5.7  (20/03/2025): AD9915: update fine output frequency setting capability.
#                      AD9912, minor: correct IFMAX and OFMAX values to avoid type
#                      checking error.
# 0.5.6  (27/02/2024): Minor: correct returned value of dds_emul.set_freq().
# 0.5.5  (24/08/2023): Enhance reliability of Ad9912Dev get/set parameters.
# 0.5.4  (23/08/2023): Handle R/W exception of FX2 device.
# 0.5.3  (04/05/2022): Correct fine tuning capability.
# 0.5.2  (08/09/2021): Add fine tuning of output frequency capability to
#                      AD9915 class and script. Enhance reliability of
#                      writing methods.
# 0.5.1  (06/09/2021): Add cli script to handle AD9915 dev board.
#                      Add explicit type conversion of ifreq, ofreq and
#                      amplitude setter method of Ad9915Dev class.
# 0.5.0  (01/09/2021): Change/Simplify name of Ad9915Dev_Umr232h to Ad9915.
#                      Enhance write method of classes Ad9912 et Ad9915.
#                      Change method name of register setter/getter of Ad9915
#                      class. Some others corrections.
# 0.4.5  (11/05/2021): Correct initialisation of ad9915dev_umr232h.
#                      Update firm_manager. Some other minor corrections.
# 0.4.4  (31/03/2021): Add emulated AD9915 development board.
# 0.4.3  (07/01/2021): Add support of AD9915 development board.
# 0.4.2  (02/07/2020): Correct import error in ad9912dev module.
# 0.4.1  (02/07/2020): Add selection of DDS device through usb bus and
#                      usb address in addition to vid and pid parameters.
# 0.4.0  (13/12/2019): Add support of signal/slot facilities in pure python
#                      using signalslot package.
# 0.3.2  (27/09/2019): Update support of pyusb to v1.0.2 instead of v1.0.0a3.
# 0.3.1  (27/06/2019): Update package name (dds -> ad9xdds).
# 0.3.0  (19/06/2019): move to PyQt5.



# -*- coding: utf-8 -*-

"""package ad9xdds
author    Benoit Dubois
copyright FEMTO ENGINEERING, 2021
license   GPL v3.0+
brief     Add signal/slot facility to Ad9915DevUmr232Hm class.
details   Class derived from Ad9915DevUmr232Hm to add signal/slot facilities.
          Signal/slot facilities use signalslot package.
          Ad9915DevUmr232Hm class allow handling of AD9915 development board
          through USB to SPI adapter.
"""

import signalslot as ss
import ad9xdds.ad9915dev_umr232hm as ad9915dev_umr232hm


class SAd9915Dev(ad9915dev_umr232hm.Ad9915Dev):
    """Class derived from Ad9915Dev class to add signal/slot facilities.
    """

    def __init__(self, **kwargs):
        self.ifreq_updated = ss.Signal(['value'])
        self.ofreq_updated = ss.Signal(['value'])
        self.phase_updated = ss.Signal(['value'])
        self.amp_updated = ss.Signal(['value'])
        super().__init__(**kwargs)

    def connect(self, url, **kwargs):
        return super().connect(url, **kwargs)

    def set_ifreq(self, value, **kwargs):
        super().set_ifreq(value)
        self.ifreq_updated.emit(value=value)

    def set_ofreq(self, value, **kwargs):
        aofreq = super().set_ofreq(value)
        self.ofreq_updated.emit(value=aofreq)
        return aofreq

    def set_phy(self, value, **kwargs):
        aphy = super().set_phy(value)
        self.phase_updated.emit(value=aphy)
        return aphy

    def set_amp(self, value, **kwargs):
        aamp = super().set_amp(value)
        self.amp_updated.emit(value=aamp)
        return aamp



import pytest
from ad9xdds.ad9915dev_umr232hm import Ad9915Dev


@pytest.mark.parametrize("ifreq, ofreq, expected", [
(1e9, 300e6, (1_288_490_188, 4, 5)),
(1e9, 260e6, (1_116_691_496, 24, 25)),
(2.5e9, 10e6, (17_179_869, 23, 125)),
])
def test_compute_modulus_parameters_basic(ifreq: float, ofreq: float, expected: list[int, int, int]):
    assert Ad9915Dev().compute_modulus_parameters(ifreq, ofreq) == expected


@pytest.mark.parametrize("ifreq, ftw, a, b, expected", [
(1e9, 1_288_490_188, 4, 5, 300e6),
(1e9, 1_116_691_496, 24, 25, 260e6),
])
def test__actual_ofreq_fine_basic(ifreq, ftw, a, b, expected):
    assert Ad9915Dev()._actual_ofreq_fine(ifreq, ftw, a, b) == expected


def test_compute_modulus_parameters_monoticity():
    ifreq = 2.5e9
    ofmin = 10e6
    ofmax = ofmin + 1e-6
    step = 1e-9

    _ofmin = int(ofmin / step)
    _ofmax = int(ofmax / step)
    ftw0 = 0
    ofreq0 = 0
    for ofreq in range(_ofmin, _ofmax):
        ofreq = ofreq * step
        ftw, a, b = Ad9915Dev().compute_modulus_parameters(ifreq, ofreq)
        assert ftw0 <= ftw 
        assert 0 <= a
        assert 0 < b
        ftw0 = ftw
        assert ofreq0 <= Ad9915Dev()._actual_ofreq_fine(ifreq, ftw, a, b), f"Monoticity error (ifreq, ftw, a, b): ({ifreq}, {ftw}, {a}, {b})"
        ofreq0 = ofreq


# Set __version__ in the setup.py
with open('ad9xdds/version.py') as f: exec(f.read())

from setuptools import setup

setup(name='ad9xdds',
      description='Ad9xDds is library dedicated to handle AD9854, AD9912 and AD9915 DDS development board.',
      version=__version__,
      packages=['ad9xdds'],
      scripts=["bin/ad9915-cli"],
      install_requires=['iopy>=0.2.0',
                        'pyftdi'],
      extras_require = {
          'Pure_Python_signalslot_facilities': ["signalslot"],
          'PyQt_signalslot_facilities': ["PyQt5"],
          'AD9854_support': ["pyparallel"]
      },
      url='https://gitlab.com/bendub/ad9xdds',
      author='Benoit Dubois',
      author_email='benoit.dubois@femto-enginneering.fr',
      license = "LGPL-3.0-or-later",
      classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python',
        'Topic :: Scientific/Engineering']
)



# -*- coding: utf-8 -*-

"""package ad9xdds
author    Benoit Dubois
copyright FEMTO ENGINEERING, 2018
license   GPL v3.0+
brief     API to control various DDS.
details   Interface that must be used to handle DDS-like device.
          The interface represent minimum method that must be implemented
          to be used in the package.
"""


# =============================================================================
def split_len(seq, length):
    """Take data of sequence type and return a list containing the n-sized
    pieces of the string.
    For example: split_len('aabbccdd', 2) => ['aa', 'bb', 'cc', 'dd']
    arg    seq     Any sequences that support slicing and len().
    arg    length  Size of slice (int).
    return         Splited sequence (list of str).
    """
    return [seq[i:i+length] for i in range(0, len(seq), length)]


# =============================================================================
def bound_value(value, vmin, vmax):
    """Check that a value is included in the range [min, max], if not the value
    is bounded to the range, ie:
    - if value < min  ->  min = value
    - if value > max  ->  max = value
    arg    value Value that is checked
    arg    vmin  Minimum valid value.
    arg    vmax  Maximum valid value.
    return       Bounded value.
    """
    if value < vmin:

        return vmin
    elif value > vmax:
        return vmax
    else:
        return value


# =============================================================================
class AbstractDds():
    """Abstract class of DDS classes. Define common properties and method for
    inherited classes.
    """

    def set_ifreq(self, ifreq):
        """Set input frequency.
        :param ifreq: Input frequency value (float)
        :returns: None
        """
        pass

    def get_ifreq(self):
        """Get input frequency.
        :returns: Current input frequency value (float)
        """
        pass

    def set_amp(self, fsc):
        """Set amplitude tuning word of output signal on DDS.
        Take the input and output frequency as argument and set the adequat
        register in the DDS.
        :param fsc: Output amplitude (int).
        :returns: fsc register value if transfert is ok (int).
        """
        pass

    def get_amp(self):
        """Get output amplitude tuning word of DDS.
        :returns: Output amplitude tuning of DDS (float).
        """
        pass

    def set_phy(self, phy):
        """Set phase of output signal on DDS.
        Take the queried output phase (in degree) as argument and set
        the adequat register in the DDS.
        :param phy:  Output phase (float).
        :returns: Actual output phase (float)
        """
        pass

    def get_phy(self):
        """Get output phase of DDS.
        :returns:  Output phase of DDS (float).
        """
        pass

    @staticmethod
    def _actual_ofreq(ifreq, ftw, rsize):
        """Return the actual output frequency.
        Due to the resolution of the DDS, the actual output frequency
        may be (a bit) different than the queried.
        :param ftw: Frequency Tuning Word register value (int).
        :param rsize: Size of register (int).
        :returns: Actual output frequency (float).
        """
        return (ftw * ifreq) / (1 << rsize)

    @staticmethod
    def _actual_phy(dphy, bit):
        """Return the actual output phase.
        Due to the resolution of the DDS, the actual output phase
        may be (a bit) different than the queried.
        :param dphy: Phase register value (int).
        :param bit: Number of bits used for the phase resolution (int)
        :returns: Actual output phase offset in degree (float).
        """
        return float(360 * dphy) / (1 << bit)



# -*- coding: utf-8 -*-

"""package ad9xdds
author    Benoit Dubois
copyright FEMTO ENGINEERING, 2020
license   GPL v3.0+
brief     API to control AD9912 DDS development board with signals/slots
          facilities.
details   Class derived Ad9912Dev class to add signals/slots facilities
"""

import signalslot as ss
import ad9xdds.ad9912dev as ad9912dev


class SAd9912Dev(ad9912dev.Ad9912Dev):
    """Class derived from Ad9912Dev class to add signal/slot facilities.
    """

    def __init__(self, **kwargs):
        self.ifreq_updated = ss.Signal(['value'])
        self.ofreq_updated = ss.Signal(['value'])
        self.phase_updated = ss.Signal(['value'])
        self.amp_updated = ss.Signal(['value'])
        self.pll_state_updated = ss.Signal(['flag'])
        self.pll_doubler_updated = ss.Signal(['flag'])
        self.pll_factor_updated = ss.Signal(['value'])
        super().__init__(**kwargs)

    def connect(self, vendor_id=ad9912dev.VENDOR_ID,
                product_id=ad9912dev.PRODUCT_ID,
                bus=None, address=None, **kwargs):
        return super().connect(vendor_id, product_id, bus, address, **kwargs)

    def set_ifreq(self, value, **kwargs):
        super().set_ifreq(value)
        self.ifreq_updated.emit(value=value)

    def set_ofreq(self, value, **kwargs):
        aofreq = super().set_ofreq(value)
        self.ofreq_updated.emit(value=aofreq)
        return aofreq

    def set_phy(self, value, **kwargs):
        aphy = super().set_phy(value)
        self.phase_updated.emit(value=aphy)
        return aphy

    def set_amp(self, value, **kwargs):
        aamp = super().set_amp(value)
        self.amp_updated.emit(value=aamp)
        return aamp

    def set_pll_state(self, flag, **kwargs):
        super().set_pll_state(flag)
        self.pll_state_updated.emit(flag=flag)

    def set_pll_doubler_state(self, flag, **kwargs):
        super().set_pll_doubler_state(flag)
        self.pll_doubler_updated.emit(flag=flag)

    def set_pll_multiplier_factor(self, value, **kwargs):
        super().set_pll_multiplier_factor(value)
        self.pll_factor_updated.emit(value=value)
